from colors import bg_color, fg_color

from shiny import App, ui

app_ui = ui.page_fluid(
    ui.tags.style(
        f"""
        :root .bslib-sidebar-layout {{
            --bslib-sidebar-bg: {bg_color};
            --bslib-sidebar-fg: {fg_color};
        }}
        """
    ),
    # Deprecated and removed in py-shiny v1.0.0
    # ui.layout_sidebar(
    #     ui.panel_sidebar("Sidebar content - 1", id="s1"),
    #     ui.panel_main("Main content - 1", id="m1"),
    # ),
    # ui.layout_sidebar(
    #     ui.panel_sidebar("Sidebar content - 2", id="s2"),
    #     ui.panel_main("Main content - 2", id="m2"),
    #     "right",
    # ),
    ui.layout_sidebar(
        ui.sidebar("Sidebar content - 3", id="s3"),
        "Main content - 3",
        id="m3",
    ),
    ui.layout_sidebar(
        ui.sidebar("Sidebar content - 4", id="s4", position="right"),
        "Main content - 4",
        id="m4",
    ),
)

app = App(app_ui, None)



# Generated by scripts/htmlDependencies.R: do not edit by hand

from __future__ import annotations

from typing import Literal

BusySpinnerType = Literal[
    "bars",
    "bars2",
    "bars3",
    "dots",
    "dots2",
    "dots3",
    "pulse",
    "pulse2",
    "pulse3",
    "ring",
    "ring2",
    "ring3",
]



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_tooltip_test_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    tooltip_auto = controller.Tooltip(page, "default_tooltip_auto")
    tooltip_auto.expect_active(False)
    tooltip_auto.set(True)
    tooltip_auto.expect_active(True)
    tooltip_auto.expect_body("An auto message")
    tooltip_auto.expect_placement(
        "right"
    )  # since there is no space on the top, it defaults to right

    tooltip_top = controller.Tooltip(page, "default_tooltip_top")
    tooltip_top.expect_active(False)
    tooltip_top.set(True)
    tooltip_top.expect_active(True)
    tooltip_top.expect_body("A top message")
    tooltip_top.expect_placement("top")

    tooltip_right = controller.Tooltip(page, "default_tooltip_right")
    tooltip_right.expect_active(False)
    tooltip_right.set(True)
    tooltip_right.expect_active(True)
    tooltip_right.expect_body("A right message")
    tooltip_right.expect_placement("right")

    tooltip_left = controller.Tooltip(page, "default_tooltip_left")
    tooltip_left.expect_active(False)
    tooltip_left.set(True)
    tooltip_left.expect_active(True)
    tooltip_left.expect_body("A left message")
    tooltip_left.expect_placement("left")



from pathlib import Path

from shiny import ui

my_theme = (
    ui.Theme("shiny", include_paths=Path(__file__).parent)
    .add_defaults(
        bslib_dashboard_design=True,
    )
    .add_mixins(
        headings_color="$success",
        bar_color="$purple",
        select_color_text="$orange",
    )
    .add_rules(
        """
        em { color: $warning; }
        .sidebar-title { color: $danger; }
        """
    )
    .add_rules('@import "css/rules.scss";')
)


if False:
    # To avoid runtime Sass compilation, save your theme CSS to a file
    # and then use that CSS file in the `theme` argument.
    css_dir = Path(__file__).parent / "css"
    css_dir.mkdir(exist_ok=True)
    with open(css_dir / "shiny-theme-demo.css", "w") as f:
        f.write(my_theme.to_css())

filler_text = """
**AI-generated filler text.** In the world of exotic fruits, the durian stands out with its spiky exterior and strong odor. Despite its divisive smell, many people are drawn to its rich, creamy texture and unique flavor profile. This tropical fruit is often referred to as the "king of fruits" in various Southeast Asian countries.

Durians are known for their large size and thorn-covered husk, _which requires careful handling_. The flesh inside can vary in color from pale yellow to deep orange, with a custard-like consistency that melts in your mouth. Some describe its taste as a mix of sweet, savory, and creamy, while others find it overpowering and pungent.
"""



from shiny.express import ui

ui.page_opts(title="App with Navbar", fillable=True, id="page")

with ui.sidebar():
    ui.input_select("data", "Dataset", ("tips", "flights", "exercise"))

    with ui.panel_conditional("input.page === 'View'"):
        ui.input_select("view", "View", ("plot", "table"))

ui.nav_spacer()

with ui.nav_panel("Data"):
    "This page could be used to pick a dataset."

with ui.nav_panel("View"):
    "This page could be used to view the dataset."
    "Notice the additional controls that appear when 'View' is selected."



import faicons as fa
import plotly.express as px

# Load data and compute static values
from shared import app_dir, tips
from shinywidgets import output_widget, render_plotly

from shiny import App, reactive, render, ui

bill_rng = (min(tips.total_bill), max(tips.total_bill))


ICONS = {
    "user": fa.icon_svg("user", "regular"),
    "wallet": fa.icon_svg("wallet"),
    "currency-dollar": fa.icon_svg("dollar-sign"),
    "ellipsis": fa.icon_svg("ellipsis"),
}

# Add page title and sidebar
app_ui = ui.page_sidebar(
    ui.sidebar(
        ui.input_slider(
            "total_bill",
            "Bill amount",
            min=bill_rng[0],
            max=bill_rng[1],
            value=bill_rng,
            pre="$",
        ),
        ui.input_checkbox_group(
            "time",
            "Food service",
            ["Lunch", "Dinner"],
            selected=["Lunch", "Dinner"],
            inline=True,
        ),
        ui.input_action_button("reset", "Reset filter"),
        open="desktop",
    ),
    ui.layout_columns(
        ui.value_box(
            "Total tippers", ui.output_ui("total_tippers"), showcase=ICONS["user"]
        ),
        ui.value_box(
            "Average tip", ui.output_ui("average_tip"), showcase=ICONS["wallet"]
        ),
        ui.value_box(
            "Average bill",
            ui.output_ui("average_bill"),
            showcase=ICONS["currency-dollar"],
        ),
        fill=False,
    ),
    ui.layout_columns(
        ui.card(
            ui.card_header("Tips data"), ui.output_data_frame("table"), full_screen=True
        ),
        ui.card(
            ui.card_header(
                "Total bill vs tip",
                ui.popover(
                    ICONS["ellipsis"],
                    ui.input_radio_buttons(
                        "scatter_color",
                        None,
                        ["none", "sex", "smoker", "day", "time"],
                        inline=True,
                    ),
                    title="Add a color variable",
                    placement="top",
                ),
                class_="d-flex justify-content-between align-items-center",
            ),
            output_widget("scatterplot"),
            full_screen=True,
        ),
        ui.card(
            ui.card_header(
                "Tip percentages",
                ui.popover(
                    ICONS["ellipsis"],
                    ui.input_radio_buttons(
                        "tip_perc_y",
                        "Split by:",
                        ["sex", "smoker", "day", "time"],
                        selected="day",
                        inline=True,
                    ),
                    title="Add a color variable",
                ),
                class_="d-flex justify-content-between align-items-center",
            ),
            output_widget("tip_perc"),
            full_screen=True,
        ),
        col_widths=[6, 6, 12],
    ),
    ui.include_css(app_dir / "styles.css"),
    title="Restaurant tipping",
    fillable=True,
)


def server(input, output, session):
    @reactive.calc
    def tips_data():
        bill = input.total_bill()
        idx1 = tips.total_bill.between(bill[0], bill[1])
        idx2 = tips.time.isin(input.time())
        return tips[idx1 & idx2]

    @render.ui
    def total_tippers():
        return tips_data().shape[0]

    @render.ui
    def average_tip():
        d = tips_data()
        if d.shape[0] > 0:
            perc = d.tip / d.total_bill
            return f"{perc.mean():.1%}"

    @render.ui
    def average_bill():
        d = tips_data()
        if d.shape[0] > 0:
            bill = d.total_bill.mean()
            return f"${bill:.2f}"

    @render.data_frame
    def table():
        return render.DataGrid(tips_data())

    @render_plotly
    def scatterplot():
        color = input.scatter_color()
        return px.scatter(
            tips_data(),
            x="total_bill",
            y="tip",
            color=None if color == "none" else color,
            trendline="lowess",
        )

    @render_plotly
    def tip_perc():
        from ridgeplot import ridgeplot

        dat = tips_data()
        dat["percent"] = dat.tip / dat.total_bill
        yvar = input.tip_perc_y()
        uvals = dat[yvar].unique()

        samples = [[dat.percent[dat[yvar] == val]] for val in uvals]

        plt = ridgeplot(
            samples=samples,
            labels=uvals,
            bandwidth=0.01,
            colorscale="viridis",
            colormode="row-index",
        )

        plt.update_layout(
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="center", x=0.5
            )
        )

        return plt

    @reactive.effect
    @reactive.event(input.reset)
    def _():
        ui.update_slider("total_bill", value=bill_rng)
        ui.update_checkbox_group("time", selected=["Lunch", "Dinner"])


app = App(app_ui, server)



from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_selectize_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    state1 = controller.InputSelectize(page, "state1")
    state2 = controller.InputSelectize(page, "state2")
    state3 = controller.InputSelectize(page, "state3")
    state4 = controller.InputSelectize(page, "state4")

    value1 = controller.OutputCode(page, "value1")
    value2 = controller.OutputCode(page, "value2")
    value3 = controller.OutputCode(page, "value3")
    value4 = controller.OutputCode(page, "value4")

    # -------------------------

    expect(state1.loc_label).to_have_text("Choose a state:")
    state1.expect_label("Choose a state:")

    state1.expect_choices(["NY", "NJ", "CT", "WA", "OR", "CA", "MN", "WI", "IA"])
    state1.expect_choice_labels(
        [
            "New York",
            "New Jersey",
            "Connecticut",
            "Washington",
            "Oregon",
            "California",
            "Minnesota",
            "Wisconsin",
            "Iowa",
        ]
    )
    state1.expect_choice_groups(["East Coast", "West Coast", "Midwest"])

    state1.expect_multiple(True)

    state1.set(["IA", "CA"])

    state1.expect_selected(["IA", "CA"])

    value1.expect_value("('IA', 'CA')")

    # -------------------------

    state2.expect_label("Selectize Options")

    state2.expect_choices(["NY", "NJ", "CT", "WA", "OR", "CA", "MN", "WI", "IA"])
    state2.expect_choice_labels(
        [
            "Select New York",
            "Select New Jersey",
            "Select Connecticut",
            "Select Washington",
            "Select Oregon",
            "Select California",
            "Select Minnesota",
            "Select Wisconsin",
            "Select Iowa",
        ]
    )
    state2.expect_choice_groups(["East Coast", "West Coast", "Midwest"])

    state2.expect_multiple(True)

    state2.set(["IA", "CA"])

    state2.expect_selected(["IA", "CA"])
    value2.expect_value("('IA', 'CA')")

    # -------------------------

    state3.expect_label("Single Selectize")

    state3.expect_choices(["NY", "NJ", "CT"])

    state3.expect_choice_labels(
        [
            "New York",
            "New Jersey",
            "Connecticut",
        ]
    )

    state3.expect_multiple(False)

    state3.set(["NJ"])

    state3.expect_selected(["NJ"])
    value3.expect_value("NJ")

    # -------------------------

    state4.expect_label("Simple Selectize")

    state4.expect_choices(["New York", "New Jersey", "Connecticut"])

    state4.expect_choice_labels(
        [
            "New York",
            "New Jersey",
            "Connecticut",
        ]
    )

    state4.expect_multiple(False)

    state4.set(["New York"])

    state4.expect_selected(["New York"])
    value4.expect_value("New York")



import pytest
from playwright.sync_api import Page
from utils.deploy_utils import reruns, reruns_delay, skip_if_not_chrome

from shiny.playwright import controller
from shiny.run import ShinyAppProc


# Edit mode becomes flaky near end of test on CI on webkit.
@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_validate_row_selection_in_edit_mode(
    page: Page, local_app: ShinyAppProc
) -> None:
    page.set_viewport_size({"width": 1920 * 2, "height": 1080 * 2})
    page.goto(local_app.url)

    # Select (and verify) a row. Edit a cell content in that row.
    # Verify the row is not focused. Hit escape key. Verify the cell value is not updated.
    # Verify the row is focused. Hit escape key again.
    # Verify the row is not focused. (Possibly verify the container div is focused?)
    data_frame = controller.OutputDataFrame(page, "penguins_df")

    data_frame.expect_cell("N1A2", row=1, col=6)
    data_frame._edit_cell_no_save("N2A2", row=1, col=6)
    data_frame._expect_row_focus_state(False, row=1)
    data_frame.expect_class_state("editing", row=1, col=6)
    data_frame.expect_selected_num_rows(1)
    data_frame.expect_selected_rows([1])
    data_frame.set_cell("N3A2", row=1, col=6, finish_key="Escape")
    data_frame.expect_cell("N1A2", row=1, col=6)
    data_frame._expect_row_focus_state(True, row=1)
    page.keyboard.press("Escape")
    data_frame._expect_row_focus_state(False, row=1)

    # Enable rows selection and editable.
    # Select (and verify) a row. Edit a cell content in that row.
    # Click a cell in another row. Verify the new row is selected and focused.
    # Verify the old row is not selected. Verify the old row cell value was updated.
    data_frame.expect_cell("N1A2", row=1, col=6)
    data_frame._edit_cell_no_save("N2A2", row=1, col=6)
    data_frame._expect_row_focus_state(False, row=1)
    data_frame.expect_class_state("editing", row=1, col=6)
    data_frame.cell_locator(row=2, col=6).click()
    data_frame._expect_row_focus_state(True, row=2)
    data_frame._expect_row_focus_state(False, row=1)
    data_frame.expect_cell("N2A2", row=1, col=6)

    # Enable rows selection and editable.
    # Select (and verify) a row. Hit enter to edit the first cell in that row.
    # Hit escape key. Verify the same row is focused.
    # Scroll right and display an html column in the left part of the view.
    # Hit enter to edit the first visible non-html cell in that row.
    # Verify that cell is editing.
    data_frame.cell_locator(row=1, col=2).click()
    page.keyboard.press("Enter")
    data_frame._expect_row_focus_state(False, row=1)
    page.keyboard.press("Escape")
    data_frame._expect_row_focus_state(True, row=1)
    page.keyboard.press("Escape")
    data_frame._edit_cell_no_save("Temp value", row=1, col=16)
    page.keyboard.press("Escape")
    page.keyboard.press("Enter")
    data_frame.expect_class_state(
        "editing",
        row=1,
        col=0,
    )  # Stage column begins to be edited.

    # Click outside the table/Press Escape to exit row focus.
    # Tab to the column name, hit enter. Verify the table becomes sorted.
    # Tab to an HTML column name, hit enter. Verify the sort does not update.
    page.keyboard.press("Escape")
    page.keyboard.press("Escape")
    page.keyboard.press("Tab")
    page.keyboard.press("Tab")  # tab to sample number
    page.keyboard.press("Enter")
    data_frame.expect_cell("152", row=0, col=1)
    page.keyboard.press("Tab")
    page.keyboard.press("Enter")
    data_frame.expect_cell("Adelie Penguin (Pygoscelis adeliae)", row=0, col=2)



from datetime import datetime

from starlette.requests import Request

from shiny import App, Inputs, Outputs, Session, reactive, render, ui


def app_ui(request: Request):
    return ui.page_fluid(
        ui.div("This page was rendered at ", datetime.now().isoformat()),
        ui.output_text("now"),
    )


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def now():
        reactive.invalidate_later(0.1)
        return f"The current time is {datetime.now().isoformat()}"


app = App(app_ui, server)



import pytest

from shiny.reactive import isolate
from shiny.render._data_frame_utils._reactive_method import reactive_calc_method


def test_is_class_method():

    class Placeholder:

        name: str

        @reactive_calc_method
        def get_name(self):
            return self.name

    ex_name = "42"
    with isolate():
        obj = Placeholder()
        obj.name = ex_name

        assert obj.get_name() == ex_name

    with pytest.raises(TypeError) as e:

        @reactive_calc_method
        def test_fn(x: object):
            return "test_fn"

    assert "reactive_calc_method" in str(e.value)

    def outer_fn():
        @reactive_calc_method
        def test_fn(x: object):
            return "test_fn"

        return test_fn

    with pytest.raises(TypeError) as e:
        outer_fn()

    def outer_placeholder():
        class InnerPlaceholder:

            name: str

            @reactive_calc_method
            def get_name(self):
                return self.name

        return InnerPlaceholder

    inner_placeholder = outer_placeholder()
    inner_placeholder.name = ex_name
    with isolate():
        obj = inner_placeholder()
        assert obj.get_name() == ex_name

    assert "<locals>" in obj.get_name.__qualname__



from icons import piggy_bank

from shiny.express import ui

with ui.layout_columns():
    with ui.value_box(
        showcase=piggy_bank, theme="bg-gradient-orange-red", full_screen=True
    ):
        "KPI Title"
        "$1 Billion Dollars"
        "Up 30% VS PREVIOUS 30 DAYS"

    with ui.value_box(
        showcase=piggy_bank,
        theme="text-green",
        showcase_layout="top right",
        full_screen=True,
    ):
        "KPI Title"
        "$1 Billion Dollars"
        "Up 30% VS PREVIOUS 30 DAYS"

    with ui.value_box(
        showcase=piggy_bank, theme="purple", showcase_layout="bottom", full_screen=True
    ):
        "KPI Title"
        "$1 Billion Dollars"
        "Up 30% VS PREVIOUS 30 DAYS"



import pytest
from playwright.sync_api import Page, expect
from utils.deploy_utils import (
    local_deploys_app_url_fixture,
    reruns,
    reruns_delay,
    skip_if_not_chrome,
)

TIMEOUT = 2 * 60 * 1000
app_url = local_deploys_app_url_fixture("example_deploy_app_a1")


@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_deploys(page: Page, app_url: str) -> None:
    page.goto(app_url)

    COUNTRY = "Afghanistan"
    expect(page.get_by_text(COUNTRY)).to_have_count(1, timeout=TIMEOUT)
    page.get_by_role("cell", name=COUNTRY).click()
    expect(page.locator("#country_detail_pop")).to_contain_text(
        COUNTRY, timeout=TIMEOUT
    )
    expect(page.locator("#country_detail_percap")).to_contain_text(COUNTRY)
    expect(page.get_by_text(COUNTRY)).to_have_count(3)



from __future__ import annotations

import functools
import os
from operator import eq
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Optional, TypeVar, cast

from .. import _utils, reactive
from .._docstring import add_example
from ..types import MISSING, MISSING_TYPE

if TYPE_CHECKING:
    from .. import Session

__all__ = ("poll", "file_reader")

T = TypeVar("T")


@add_example()
def poll(
    poll_func: Callable[[], Any] | Callable[[], Awaitable[Any]],
    interval_secs: float = 1,
    *,
    equals: Callable[[Any, Any], bool] = eq,
    priority: int = 0,
    session: MISSING_TYPE | Session | None = MISSING,
) -> Callable[[Callable[[], T]], Callable[[], T]]:
    """
    Create a reactive polling object.

    Polling is a technique that approximates "real-time" or streaming updates, as if a
    data source were pushing notifications each time it is updated. The data source does
    not actually push notifications; a polling object repeatedly checks for changes in an
    efficient way at specified intervals. If a change is detected, the polling object runs
    a function to re-read the data source.

    A reactive polling object is constructed using two functions: a polling function,
    which is a fast-running, inexpensive function that is used to determine whether some
    data source has changed (such as the timestamp of a file, or a `SELECT MAX(updated)
    FROM table` query); and a slower-running reading function that actually loads and
    returns the data that is desired. The `poll()` function is intended to be used as a
    decorator: the poll function is passed as the `poll_func` arg to `@poll()`, while
    the data reading function is the target of the decorator.

    Reactive consumers can invoke the resulting polling object to get the current data,
    and will automatically invalidate when the polling function detects a change.
    Polling objects also cache the results of the read function; for this reason, apps
    where all sessions depend on the same data source may want to declare the polling
    object at the top level of app.py (outside of the server function).

    Both `poll_func` and the decorated (data reading) function can read reactive values
    and :func:`~shiny.reactive.calc` objects. Any invalidations triggered by reactive
    dependencies will apply to the reactive polling object immediately (not waiting for
    the `interval_secs` delay to expire).

    Parameters
    ----------
    poll_func
        A function to be called frequently to determine whether a data source has
        changed. The return value should be something that can be compared inexpensively
        using `==`. Both regular functions and co-routine functions are allowed.

        Note that the `poll_func` should NOT return a bool that indicates whether the
        data source has changed. Rather, each `poll_func` return value will be checked
        for equality with its preceding `poll_func` return value (using `==` semantics
        by default), and if it differs, the data source will be considered changed.
    interval_secs
        The number of seconds to wait after each `poll_func` invocation before polling
        again. Note: depending on what other tasks are executing, the actual wait time
        may far exceed this value.
    equals
        The function that will be used to compare each `poll_func` return value with its
        immediate predecessor.
    priority
        Reactive polling is implemented using an :func:`~shiny.reactive.effect` to call
        `poll_func` on a timer; use the `priority` argument to control the order of this
        Effect's execution versus other Effects in your app. See
        :func:`~shiny.reactive.effect` for more details.
    session
        A :class:`~shiny.Session` instance. If not provided, a session is inferred via
        :func:`~shiny.session.get_current_session`. If there is no current session (i.e.
        `poll` is being created outside of the server function), the lifetime of this
        reactive poll object will not be tied to any specific session.

    Returns
    -------
    :
        A decorator that should be applied to a no-argument function that (expensively)
        reads whatever data is desired. (This function may be a regular function or a
        co-routine function.) The result of the decorator is a reactive
        :func:`~shiny.reactive.calc` that always returns up-to-date data, and
        invalidates callers when changes are detected via polling.

    See Also
    --------
    * :func:`~shiny.reactive.file_reader`
    """

    with reactive.isolate():
        last_value: reactive.Value[Any] = reactive.Value(poll_func())
        last_error: reactive.Value[Optional[Exception]] = reactive.Value(None)

    @reactive.effect(priority=priority, session=session)
    async def _():
        try:
            if _utils.is_async_callable(poll_func):
                new = await poll_func()
            else:
                new = poll_func()

            with reactive.isolate():
                old = last_value.get()

            try:
                is_equal = equals(old, new)
            except Exception as e:
                # For example, pandas DataFrame throws if you try to compare it to a
                # non-comparable object
                raise TypeError(
                    "The reactive.poll polling function returned an object that "
                    "couldn't be compared with a previously returned object. Try "
                    "modifying your polling function to return a simpler type, like a "
                    "str, float, list, or dict."
                ) from e
            if not isinstance(is_equal, bool):
                # Comparison succeeded, but we don't understand the result
                if equals is eq:
                    # We used == but it didn't work
                    raise TypeError(
                        "The reactive.poll polling function returned an object "
                        "that doesn't implement a simple == operator. Try "
                        "modifying your polling function to return a simpler type, "
                        "like a str, float, list, or dict."
                    )
                else:
                    # The caller passed in a custom function
                    raise TypeError(
                        "The reactive.poll `equals` function returned a non-bool "
                        "value"
                    )

            # If we got here, the comparison succeeded. Need to make sure the error is
            # cleared, but don't unnecessarily call last_error.set(); at the time of
            # this writing, we haven't made a final decision on whether reactive.Value
            # will ignore sets if the new value is identical to the existing one.
            with reactive.isolate():
                if last_error.get() is not None:
                    last_error.set(None)

            if not is_equal:
                last_value.set(new)
        except Exception as e:
            # Either the polling function threw an error, or we failed to compare its
            # result with a previous result. Either way, we failed; save the error so
            # that it can be exposed to whoever's trying to use the poll object.
            last_error.set(e)
        finally:
            reactive.invalidate_later(interval_secs)

    def wrapper(fn: Callable[[], T]) -> Callable[[], T]:
        if _utils.is_async_callable(fn):

            @reactive.calc(session=session)
            @functools.wraps(fn)
            async def result_async() -> T:
                # If an error occurred, raise it
                err = last_error.get()
                if err is not None:
                    raise err

                # Take dependency on polling result
                last_value.get()

                # Note that we also depend on the main function
                return await fn()

            # In this code path, the cast is necessary because result_async() has an
            # incorrect signature due to limitations in Python's type hints. In this
            # path, T is already an Awaitable. The signature should really be `async def
            # result_async() -> Awaited[T]`, but there's no Awaited type in Python. So
            # instead we'll just cast() it.
            return cast(Callable[[], T], result_async)

        else:

            @reactive.calc(session=session)
            @functools.wraps(fn)
            def result_sync() -> T:
                # If an error occurred, raise it
                err = last_error.get()
                if err is not None:
                    raise err

                # Take dependency on polling result
                last_value.get()

                # Note that we also depend on the main function
                return fn()

            return result_sync

    return wrapper


@add_example()
def file_reader(
    filepath: (
        str | os.PathLike[str] | Callable[[], str] | Callable[[], os.PathLike[str]]
    ),
    interval_secs: float = 1,
    *,
    priority: int = 1,
    session: MISSING_TYPE | Session | None = MISSING,
) -> Callable[[Callable[[], T]], Callable[[], T]]:
    """
    Create a reactive file reader.

    This is a decorator, meant to be applied to a no-argument function that reads data
    from a file on disk. Whenever the file changes (or to be precise, the file size or
    last modified time changes), past readers of the data are reactively invalidated.
    This makes it incredibly easy to write apps that automatically update all of their
    outputs as soon as files on disk change.

    Note that `file_reader` works only on single files, not directories of files.

    Both the `filepath` function and the decorated (file reading) function can read
    reactive values and :func:`~shiny.reactive.calc` objects. Any invalidations
    triggered by reactive dependencies will apply to the reactive file reader object
    immediately (not waiting for the `interval_secs` delay to expire).

    Parameters
    ----------
    filepath
        Either a string indicating the file path to be monitored, or, a no-argument
        function that returns such a string. The latter is useful if the file to be
        monitored depends on some user input, the current date/time, etc.

        The file path provided MUST exist, otherwise Shiny will treat it as an unhandled
        error and close the session.

        If a function is used, make sure it is high performance (or is cached, i.e. use
        a :class:`~shiny.reactive.calc)`, as it will be called very frequently.
    interval_secs
        The number of seconds to wait after each time the file metadata is checked.
        Note: depending on what other tasks are executing, the actual wait time may far
        exceed this value.
    priority
        Reactive polling is implemented using an :func:`~shiny.reactive.effect` to call
        `poll_func` on a timer; use the `priority` argument to control the order of this
        Effect's execution versus other Effects in your app. See
        :func:`~shiny.reactive.effect` for more details.
    session
        A :class:`~shiny.Session` instance. If not provided, a session is inferred via
        :func:`~shiny.session.get_current_session`. If there is no current session (i.e.
        `poll` is being created outside of the server function), the lifetime of this
        reactive poll object will not be tied to any specific session.

    Returns
    -------
    :
        A decorator that should be applied to a no-argument function that (expensively)
    reads whatever data is desired. (This function may be a regular function or a
    co-routine function.) The result of the decorator is a reactive
    :func:`~shiny.reactive.calc` that always returns up-to-date data, and invalidates
    callers when changes are detected via polling.

    See Also
    --------
    * :func:`~shiny.reactive.poll`
    """

    if isinstance(filepath, str):
        # Normalize filepath so it's always a function

        filepath_value = filepath

        def filepath_func_str() -> str:
            return filepath_value

        filepath = filepath_func_str
    elif isinstance(filepath, os.PathLike):
        filepath_value = filepath

        def filepath_func_pathlike() -> os.PathLike[str]:
            return filepath_value

        filepath = filepath_func_pathlike

    def check_timestamp():
        path = filepath()
        return (path, os.path.getmtime(path), os.path.getsize(path))

    def wrapper(fn: Callable[[], T]) -> Callable[[], T]:
        if _utils.is_async_callable(fn):

            @poll(
                check_timestamp,
                interval_secs=interval_secs,
                priority=priority,
                session=session,
            )
            @functools.wraps(fn)
            async def reader_async():
                return await fn()

            # See poll() for explanation of why this cast is needed.
            return cast(Callable[[], T], reader_async)
        else:

            @poll(
                check_timestamp,
                interval_secs=interval_secs,
                priority=priority,
                session=session,
            )
            @functools.wraps(fn)
            def reader():
                return fn()

            return reader

    return wrapper



# pyright: reportUnknownParameterType=false
# pyright: reportUnknownMemberType=false
# pyright: reportInvalidTypeForm=false
# pyright: reportUntypedFunctionDecorator=false
# pyright: reportUnknownArgumentType=false
# pyright: reportFunctionMemberAccess=false
from __future__ import annotations

import asyncio
from typing import Any, cast, overload

import pytest

from shiny._deprecated import ShinyDeprecationWarning
from shiny._namespaces import ResolvedId, Root
from shiny._utils import is_async_callable
from shiny.render.transformer import (
    TransformerMetadata,
    ValueFn,
    output_transformer,
    resolve_value_fn,
)
from shiny.session import Session, session_context

# import warnings
# warnings.filterwarnings("ignore", category=ShinyDeprecationWarning)


class _MockSession:
    ns: ResolvedId = Root

    # This is needed so that Outputs don't throw an error.
    def _is_hidden(self, name: str) -> bool:
        return False


test_session = cast(Session, _MockSession())


def test_output_transformer_works():
    # No args works
    @output_transformer
    async def TestTransformer(
        _meta: TransformerMetadata,
        _fn: ValueFn[str],
    ): ...

    @overload
    def test_renderer() -> TestTransformer.OutputRendererDecorator: ...

    @overload
    def test_renderer(
        _fn: TestTransformer.ValueFn,
    ) -> TestTransformer.OutputRenderer: ...

    def test_renderer(
        _fn: TestTransformer.ValueFn | None = None,
    ) -> TestTransformer.OutputRenderer | TestTransformer.OutputRendererDecorator:
        return TestTransformer(_fn)


def test_output_transformer_kwargs_are_allowed():
    # Test that kwargs can be allowed
    @output_transformer
    async def TestTransformer(
        _meta: TransformerMetadata,
        _fn: ValueFn[str],
        *,
        y: str = "42",
    ): ...

    @overload
    def test_renderer(*, y: str = "42") -> TestTransformer.OutputRendererDecorator: ...

    @overload
    def test_renderer(
        _fn: TestTransformer.ValueFn,
    ) -> TestTransformer.OutputRenderer: ...

    def test_renderer(
        _fn: TestTransformer.ValueFn | None = None,
        *,
        y: str = "42",
    ) -> TestTransformer.OutputRenderer | TestTransformer.OutputRendererDecorator:
        return TestTransformer(
            _fn,
            TestTransformer.params(y=y),
        )


def test_output_transformer_with_pass_through_kwargs():
    # No args works
    @output_transformer
    async def TestTransformer(
        _meta: TransformerMetadata,
        _fn: ValueFn[str],
        *,
        y: str = "42",
        **kwargs: float,
    ): ...

    @overload
    def test_renderer(
        *, y: str = "42", **kwargs: Any
    ) -> TestTransformer.OutputRendererDecorator: ...

    @overload
    def test_renderer(
        _fn: TestTransformer.ValueFn,
    ) -> TestTransformer.OutputRenderer: ...

    def test_renderer(
        _fn: TestTransformer.ValueFn | None = None,
        *,
        y: str = "42",
        **kwargs: Any,
    ) -> TestTransformer.OutputRenderer | TestTransformer.OutputRendererDecorator:
        return TestTransformer(
            _fn,
            TestTransformer.params(y=y, **kwargs),
        )


def test_output_transformer_pos_args():
    try:

        @output_transformer  # pyright: ignore[reportArgumentType]
        async def TestTransformer(
            _meta: TransformerMetadata,
        ): ...

        raise RuntimeError()
    except TypeError as e:
        assert "must have 2 positional parameters" in str(e)


def test_output_transformer_limits_positional_arg_count():
    try:

        @output_transformer
        async def TestTransformer(
            _meta: TransformerMetadata,
            _fn: ValueFn[str],
            y: str,
        ): ...

        raise RuntimeError()
    except TypeError as e:
        assert "more than 2 positional" in str(e)


def test_output_transformer_does_not_allow_args():
    try:

        @output_transformer
        async def TestTransformer(
            _meta: TransformerMetadata,
            _fn: ValueFn[str],
            *args: str,
        ): ...

        raise RuntimeError()

    except TypeError as e:
        assert "No variadic positional parameters" in str(e)


def test_output_transformer_kwargs_have_defaults():
    try:

        @output_transformer
        async def TestTransformer(
            _meta: TransformerMetadata,
            _fn: ValueFn[str],
            *,
            y: str,
        ): ...

        raise RuntimeError()

    except TypeError as e:
        assert "did not have a default value" in str(e)


def test_output_transformer_result_does_not_allow_args():
    @output_transformer
    async def TestTransformer(
        _meta: TransformerMetadata,
        _fn: ValueFn[str],
    ): ...

    # Test that args can **not** be supplied
    def render_fn_sync(*args: str):
        return " ".join(args)

    try:
        TestTransformer(
            render_fn_sync,
            "X",  # pyright: ignore[reportArgumentType]
        )
        raise RuntimeError()
    except TypeError as e:
        assert "Expected `params` to be of type `TransformerParams`" in str(e)


@pytest.mark.asyncio
async def test_renderer_handler_or_transform_fn_can_be_async():
    @output_transformer
    async def AsyncTransformer(
        _meta: TransformerMetadata,
        _fn: ValueFn[str],
    ) -> str:
        if is_async_callable(_fn):
            # Conditionally sleep to test that the handler is truly async
            await asyncio.sleep(0)
        ret = await resolve_value_fn(_fn)
        return ret

    # ## Setup overloads =============================================

    @overload
    def async_renderer() -> AsyncTransformer.OutputRendererDecorator: ...

    @overload
    def async_renderer(
        _fn: AsyncTransformer.ValueFn,
    ) -> AsyncTransformer.OutputRenderer: ...

    def async_renderer(
        _fn: AsyncTransformer.ValueFn | None = None,
    ) -> AsyncTransformer.OutputRenderer | AsyncTransformer.OutputRendererDecorator:
        with pytest.warns(ShinyDeprecationWarning):
            return AsyncTransformer(_fn)

    test_val = "Test: Hello World!"

    # ## Test Sync: X =============================================

    @async_renderer
    def renderer_sync() -> str:
        return test_val

    # All renderers are async in execution.
    assert not is_async_callable(renderer_sync._value_fn)

    with session_context(test_session):
        val = await renderer_sync._run()
        assert val == test_val

    # ## Test Async: √ =============================================

    async_test_val = "Async: Hello World!"

    @async_renderer
    async def renderer_async() -> str:
        await asyncio.sleep(0)
        return async_test_val

    if not is_async_callable(renderer_async._value_fn):
        raise RuntimeError("Expected `renderer_async` to be a coro function")

    with session_context(test_session):
        ret = await renderer_async._run()
        assert ret == async_test_val


# @pytest.mark.asyncio
# async def test_resolve_value_fn_is_deprecated():
#     with pytest.warns(ShinyDeprecationWarning):
#         test_val = 42

#         async def value_fn():
#             return test_val

#         ret = await resolve_value_fn(value_fn)
#         assert test_val == ret



from __future__ import annotations

import modin.pandas as md  # pyright: ignore[reportMissingTypeStubs]
import narwhals.stable.v1 as nw
import palmerpenguins  # pyright: ignore[reportMissingTypeStubs]

from shiny.express import render, ui

pd_df = palmerpenguins.load_penguins_raw().iloc[0:2, 0:2]

nw_df = nw.from_native(pd_df, eager_only=True)

md_df = md.DataFrame(pd_df)

with ui.card():

    ui.h2("Polars Pandas Data")

    @render.table
    def nw_table():
        return nw_df

    "Data type:"

    @render.code
    def nw_df_type():
        return str(type(nw_df))


with ui.card():

    ui.h2("Modin dataframe Data")

    @render.table
    def md_table():
        return md_df

    "Data type:"

    @render.code
    def md_df_type():
        return str(type(md_df))



from __future__ import annotations

import re
from typing import Literal

from .._typing_extensions import NotRequired, TypedDict


class ClickOpts(TypedDict):
    id: NotRequired[str]
    clip: bool


class DblClickOpts(TypedDict):
    id: NotRequired[str]
    clip: bool
    delay: int


class HoverOpts(TypedDict):
    id: NotRequired[str]
    delay: int
    delayType: Literal["debounce", "throttle"]
    clip: bool
    nullOutside: bool


class BrushOpts(TypedDict):
    id: NotRequired[str]
    fill: str
    stroke: str
    opacity: float
    delay: int
    delayType: Literal["debounce", "throttle"]
    clip: bool
    direction: str
    resetOnNew: bool


# It would be better if opts could just be a Dict[str, Union[str, bool]], but ClickOpts
# is a TypedDict, and that is _not_ a subclass of Dict because it doesn't support some
# Dict operations, like removing items specified in the TypedDict.
def format_opt_names(
    opts: ClickOpts | DblClickOpts | HoverOpts | BrushOpts,
    prefix: str,
) -> dict[str, str]:
    new_opts: dict[str, str] = dict()
    for key, value in opts.items():
        new_key = f"data-{prefix}-" + re.sub("([A-Z])", "-\\1", key).lower()
        new_value = str(value)

        if isinstance(value, bool):
            new_value = new_value.lower()

        new_opts[new_key] = new_value

    return new_opts


def click_opts(
    *,
    clip: bool = True,
) -> ClickOpts:
    return {
        "clip": clip,
    }


def dblclick_opts(
    *,
    delay: int = 400,
    clip: bool = True,
) -> DblClickOpts:
    return {
        "delay": delay,
        "clip": clip,
    }


def hover_opts(
    *,
    delay: int = 300,
    delay_type: Literal["debounce", "throttle"] = "debounce",
    clip: bool = True,
    null_outside: bool = True,
) -> HoverOpts:
    return {
        "delay": delay,
        "delayType": delay_type,
        "clip": clip,
        "nullOutside": null_outside,
    }


def brush_opts(
    *,
    fill: str = "#9cf",
    stroke: str = "#036",
    opacity: float = 0.25,
    delay: int = 300,
    delay_type: Literal["debounce", "throttle"] = "debounce",
    clip: bool = True,
    direction: str = "xy",
    reset_on_new: bool = False,
) -> BrushOpts:
    return {
        "fill": fill,
        "stroke": stroke,
        "opacity": opacity,
        "delay": delay,
        "delayType": delay_type,
        "clip": clip,
        "direction": direction,
        "resetOnNew": reset_on_new,
    }



from pathlib import Path

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from shiny import App, Inputs, Outputs, Session, reactive, render, ui
from shiny.plotutils import brushed_points

path = Path(__file__).parent / "boulder_temp.csv"
weather_df = pd.read_csv(path)
weather_df["date"] = pd.to_datetime(weather_df["date"])
weather_df["annotation"] = ""

app_ui = ui.page_fluid(
    ui.panel_title("Plot annotation example"),
    ui.p(
        """
        Select points to annotate them.
        The plot is rendered with seaborn and all interaction is handled by Shiny.
        """,
        {"style": "font-size: larger"},
    ),
    ui.row(
        ui.column(
            6,
            ui.output_plot("time_series", brush=ui.brush_opts(direction="x")),
            ui.output_ui("annotator"),
        ),
        ui.column(
            4,
            ui.h3("Annotated points"),
            ui.output_data_frame("annotations"),
        ),
        ui.column(2, ui.download_button("download", "Download CSV")),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    annotated_data = reactive.value(weather_df)

    @reactive.calc
    def selected_data():
        out = brushed_points(annotated_data(), input.time_series_brush(), xvar="date")
        return out

    @reactive.effect
    @reactive.event(input.annotate_button)
    def _():
        selected = selected_data()
        selected["annotation_new"] = input.annotation()
        selected = selected.loc[:, ["date", "annotation_new"]]

        df = annotated_data().copy()

        df = df.merge(selected, on="date", how="left")
        df["annotation_new"] = df["annotation_new"].fillna("")
        updated_rows = df["annotation_new"] != ""
        df.loc[updated_rows, "annotation"] = df.loc[updated_rows, "annotation_new"]
        df = df.loc[:, ["date", "temp_c", "annotation"]]
        annotated_data.set(df)

    @render.plot
    def time_series():
        fig, ax = plt.subplots()
        ax.xaxis.set_major_locator(mdates.MonthLocator())
        ax.xaxis.set_major_formatter(mdates.DateFormatter("%Y-%m"))
        ax.set_title("Temperature readings, Boulder Colorado")
        out = sns.scatterplot(
            data=annotated_data(), x="date", y="temp_c", hue="annotation", ax=ax
        )

        out.tick_params(axis="x", rotation=30)
        return out.get_figure()

    @render.ui
    def annotator():
        if input.time_series_brush() is not None:
            selected = selected_data()

            min = str(selected["date"].min())
            max = str(selected["date"].max())

            min = min.replace(" 00:00:00+00:00", "")
            max = max.replace(" 00:00:00+00:00", "")

            out = ui.TagList(
                ui.row(
                    {"style": "padding-top: 20px;"},
                    ui.column(
                        4,
                        ui.p(f"{min} to", ui.br(), f"{max}"),
                    ),
                    ui.column(
                        4,
                        ui.input_text("annotation", "", placeholder="Enter annotation"),
                    ),
                    ui.column(4, ui.input_action_button("annotate_button", "Submit")),
                )
            )
            return out

    @render.data_frame
    def annotations():
        df = annotated_data().copy()
        df["date"] = df["date"].dt.strftime("%Y-%m-%d")
        df = df.loc[df["annotation"] != ""]
        return df

    @render.download(filename="data.csv")
    def download():
        yield annotated_data().to_csv()


app = App(app_ui, server)



# ------------------------------------------------------------------------------------
# A basic Shiny Chat example powered by OpenAI via LangChain.
# To run it, you'll need OpenAI API key.
# To get one, follow the instructions at https://platform.openai.com/docs/quickstart
# To use other providers/models via LangChain, see https://python.langchain.com/v0.1/docs/modules/model_io/chat/quick_start/
# ------------------------------------------------------------------------------------
import os

from app_utils import load_dotenv
from langchain_openai import ChatOpenAI

from shiny.express import ui

# Either explicitly set the OPENAI_API_KEY environment variable before launching the
# app, or set them in a file named `.env`. The `python-dotenv` package will load `.env`
# as environment variables which can later be read by `os.getenv()`.
load_dotenv()
llm = ChatOpenAI(api_key=os.environ.get("OPENAI_API_KEY"))  # type: ignore

# Set some Shiny page options
ui.page_opts(
    title="Hello LangChain Chat Models",
    fillable=True,
    fillable_mobile=True,
)

# Create and display an empty chat UI
chat = ui.Chat(id="chat")
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    messages = chat.messages(format="langchain")
    # Create a response message stream
    response = llm.astream(messages)
    # Append the response stream into the chat
    await chat.append_message_stream(response)



import pandas as pd
import seaborn as sns

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(
    ui.head_content(
        ui.tags.meta(name="viewport", content="width=device-width, initial-scale=1")
    ),
    ui.input_select(
        "selection_mode",
        "Selection mode",
        {
            "none": "(None)",
            "row": "Single row",
            "rows": "Multiple rows",
        },
        selected="rows",
    ),
    ui.input_switch("filters", "Filters", True),
    ui.input_switch("editable", "Editable", True),
    ui.layout_column_wrap(
        ui.card(
            ui.card_header("Data Frame as ", ui.tags.code("render.DataGrid")),
            ui.output_data_frame("grid"),
        ),
        ui.card(
            ui.card_header("Data Frame as ", ui.tags.code("render.DataTable")),
            ui.output_data_frame("table"),
        ),
        width=1 / 2,
    ),
    class_="p-3",
)


def server(input: Inputs, output: Outputs, session: Session):
    df: reactive.value[pd.DataFrame] = reactive.value(
        sns.load_dataset("anagrams").iloc[:, 1:]
    )

    def update_data_with_patch(patch):
        df_copy = df().copy()
        fn = str if patch["column_index"] == 0 else int
        df_copy.iat[patch["row_index"], patch["column_index"]] = fn(patch["value"])
        df.set(df_copy)

    height = 350
    width = "fit-content"

    @render.data_frame
    def grid():
        return render.DataGrid(
            df(),
            width=width,
            height=height,
            filters=input.filters(),
            editable=input.editable(),
            selection_mode=input.selection_mode(),
        )

    @grid.set_patch_fn
    def _(*, patch: render.CellPatch):
        update_data_with_patch(patch)
        return patch["value"]

    @render.data_frame
    def table():
        return render.DataTable(
            df(),
            width=width,
            height=height,
            filters=input.filters(),
            editable=input.editable(),
            selection_mode=input.selection_mode(),
        )

    @table.set_patch_fn
    def _(*, patch: render.CellPatch):
        update_data_with_patch(patch)
        return patch["value"]


app = App(app_ui, server)



# ------------------------------------------------------------------------------------
# A basic Shiny Chat example powered by OpenAI's GPT-4o model.
# To run it, you'll need OpenAI API key.
# To get setup, follow the instructions at https://platform.openai.com/docs/quickstart
# ------------------------------------------------------------------------------------
import os

from app_utils import load_dotenv
from openai import AsyncOpenAI

from shiny.express import ui

# Either explicitly set the OPENAI_API_KEY environment variable before launching the
# app, or set them in a file named `.env`. The `python-dotenv` package will load `.env`
# as environment variables which can later be read by `os.getenv()`.
load_dotenv()
llm = AsyncOpenAI(api_key=os.environ.get("OPENAI_API_KEY"))

# Set some Shiny page options
ui.page_opts(
    title="Hello OpenAI Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create a chat instance, with an initial message
chat = ui.Chat(
    id="chat",
    messages=[
        {"content": "Hello! How can I help you today?", "role": "assistant"},
    ],
)

# Display the chat
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    messages = chat.messages(format="openai")
    # Create a response message stream
    response = await llm.chat.completions.create(
        model="gpt-4o",
        messages=messages,
        stream=True,
    )
    # Append the response stream into the chat
    await chat.append_message_stream(response)



from pathlib import PurePath

from htmltools import HTMLDependency, Tag

from shiny.module import resolve_id
from shiny.render.renderer import Jsonifiable, Renderer

# This object is used to let Shiny know where the dependencies needed to run
# our component all live. In this case, we're just using a single javascript
# file but we could also include CSS.
custom_component_deps = HTMLDependency(
    "custom_component",
    "1.0.0",
    source={
        "package": "custom_component",
        "subdir": str(PurePath(__file__).parent / "distjs"),
    },
    script={"src": "index.js", "type": "module"},
)


def input_custom_component(id: str):
    """
    A shiny input.
    """
    return Tag(
        # This is the name of the custom tag we created with our webcomponent
        "custom-component-input",
        custom_component_deps,
        # Use resolve_id so that our component will work in a module
        id=resolve_id(id),
    )


# Output component
class render_custom_component(Renderer[str]):
    """
    Render a value in a custom component.
    """

    # The UI used within Shiny Express mode
    def auto_output_ui(self) -> Tag:
        return output_custom_component(self.output_id)

    # # There are no parameters being supplied to the `output_custom_component` rendering function.
    # # Therefore, we can omit the `__init__()` method.
    # def __init__(self, _fn: Optional[ValueFn[int]] = None, *, extra_arg: str = "bar"):
    #     super().__init__(_fn)
    #     self.extra_arg: str = extra_arg

    # Transforms non-`None` values into a `Jsonifiable` object.
    # If you'd like more control on when and how the value is rendered,
    # please use the `async def render(self)` method.
    async def transform(self, value: str) -> Jsonifiable:
        # Send the results to the client. Make sure that this is a serializable
        # object and matches what is expected in the javascript code.
        return {"value": str(value)}


def output_custom_component(id: str):
    """
    Show a color
    """
    return Tag(
        "custom-component-output",
        custom_component_deps,
        id=resolve_id(id),
    )



import functools
import time
from contextlib import contextmanager
from typing import Any, Callable, Generator

from ..._docstring import no_example

__all__ = ("expect_to_change",)


@no_example()
@contextmanager
def expect_to_change(
    func: Callable[[], Any], timeout_secs: float = 10
) -> Generator[None, None, None]:
    """
    Context manager that yields when the value returned by func() changes.

    Use this around code that has a side-effect of changing some state asynchronously
    (such as all browser actions), to prevent moving onto the next step of the test
    until this one has actually taken effect.

    Parameters
    ----------
    func
        A function that returns a value. The value returned by this function is
        compared to the value returned by subsequent calls to this function.
    timeout_secs
        How long to wait for the value to change before raising TimeoutError.

    Raises
    ------
    TimeoutError
        If the value does not change within timeout_secs.

    Example
    -------

        with expect_to_change(lambda: page.locator("#name").value()):
            page.keyboard.send_keys("hello")

    """

    original_value = func()
    yield

    @retry_with_timeout(timeout_secs)
    def wait_for_change():
        if func() == original_value:
            raise AssertionError("Value did not change")

    wait_for_change()


@no_example()
def retry_with_timeout(timeout: float = 30):
    """
    Decorator that retries a function until 1) it succeeds, 2) fails with a
    non-assertion error, or 3) repeatedly fails with an AssertionError for longer than
    the timeout. If the timeout elapses, the last AssertionError is raised.

    Parameters
    ----------
    timeout
        How long to wait for the function to succeed before raising the last
        AssertionError.

    Returns
    -------
    A decorator that can be applied to a function.

    Example
    -------

        @retry_with_timeout(30)
        def try_to_find_element():
            if not page.locator("#name").exists():
                raise AssertionError("Element not found")

        try_to_find_element()
    """

    def decorator(func: Callable[[], None]) -> Callable[[], None]:
        @functools.wraps(func)
        def wrapper() -> None:
            start = time.time()
            while True:
                try:
                    return func()
                except AssertionError as e:
                    if time.time() - start > timeout:
                        raise e
                    time.sleep(0.1)

        return wrapper

    return decorator



from __future__ import annotations

# TODO-future; When `OutputRenderer` is removed, remove `output_args()`


__all__ = (
    "TransformerMetadata",
    "TransformerParams",
    "OutputRenderer",
    "OutputTransformer",
    "ValueFn",
    # "ValueFnSync",
    # "ValueFnAsync",
    # "TransformFn",
    "output_transformer",
    "is_async_callable",
)

import inspect
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Generic,
    NamedTuple,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from ..renderer import Jsonifiable, Renderer
from ..renderer._renderer import DefaultUIFn, DefaultUIFnResultOrNone

if TYPE_CHECKING:
    from ...session import Session

from ..._deprecated import warn_deprecated
from ..._docstring import add_example
from ..._typing_extensions import Concatenate, ParamSpec
from ..._utils import is_async_callable
from ...types import MISSING

# Input type for the user-spplied function that is passed to a render.xx
IT = TypeVar("IT")
# Output type after the OutputRenderer._exec method has been called on the IT object.
OT = TypeVar("OT")
# Param specification for value_fn function
P = ParamSpec("P")


# ======================================================================================
# Helper classes
# ======================================================================================


# Meta information to give `hander()` some context
class TransformerMetadata(NamedTuple):
    """
    Transformer metadata

    This class is used to hold meta information for a transformer function.

    Attributes
    ----------
    session
        The :class:`~shiny.Session` object of the current output value function.
    name
        The name of the output being rendered.
    """

    session: Session
    name: str


# Motivation for using this class:
# * https://peps.python.org/pep-0612/ does allow for prepending an arg (e.g.
#   `value_fn`).
# * However, the overload is not happy when both a positional arg (e.g. `value_fn`) is
#   dropped and the variadic positional args (`*args`) are kept.
# * The variadic positional args (`*args`) CAN NOT be dropped as PEP612 states that both
#   components of the `ParamSpec` must be used in the same function signature.
# * By making assertions on `P.args` to only allow for `*`, we _can_ make overloads that
#   use either the single positional arg (e.g. `value_fn`) or the `P.kwargs` (as
#   `P.args` == `*`)
class TransformerParams(Generic[P]):
    """
    Parameters for a transformer function

    This class is used to isolate the transformer function parameters away from
    internal implementation parameters used by Shiny.

    """

    def __init__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """
        Parameters
        ----------
        *args
            No positional arguments should be supplied. Only keyword arguments should be
            supplied. (`*args` is required when using :class:`~typing.ParamSpec` even if
            transformer is only leveraging `**kwargs`.)
        **kwargs
            Keyword arguments for the corresponding renderer function.
        """

        # Make sure there no `args` at run time!
        # This check is related to `_assert_transform_fn` not accepting any `args`
        if len(args) > 0:
            print(args)
            raise RuntimeError("`args` should not be supplied")

        # `*args` must be defined with `**kwargs` (as per PEP612)
        # (even when expanded later when calling the transform function)
        # We need to store (and later retrieve) them, even if we know they are empty
        self.args = args
        self.kwargs = kwargs


# ======================================================================================
# Renderer / RendererSync / RendererAsync base class
# ======================================================================================

# A `ValueFn` function is an app-supplied function which returns an IT.
# It can be either synchronous or asynchronous
ValueFnSync = Callable[[], IT]
"""
App-supplied output value function which returns type `IT`. This function is
synchronous.
"""
ValueFnAsync = Callable[[], Awaitable[IT]]
"""
App-supplied output value function which returns type `IT`. This function is
asynchronous.
"""

# ValueFnAsync = ValueFnAsync[IT]
# """
# App-supplied output value function which returns type `IT`. This function is always
# asyncronous as the original app-supplied function possibly wrapped to execute
# asynchonously.
# """

ValueFn = Union[ValueFnSync[IT], ValueFnAsync[IT]]
"""
App-supplied output value function which returns type `IT`. This function can be
synchronous or asynchronous.
"""

# `TransformFn` is a package author function that transforms an object of type `IT` into
# type `OT`.
TransformFn = Callable[Concatenate[TransformerMetadata, ValueFn[IT], P], Awaitable[OT]]
"""
Package author function that transforms an object of type `IT` into type `OT`.
"""


class OutputRenderer(
    # Inherit from `[Any]` to not change the type signature of `OutputRenderer[OT]` to `OuputRenderer[IT, OT]`.
    Renderer[Any],
    Generic[OT],
):
    """
    Output Renderer

    Transforms the output (of type `IT`) of an app-supplied output value function
    (`value_fn`) into type (`OT`). This transformed value is then sent to be an
    :class:`~shiny.Outputs` output value.

    When the `._run` method is invoked, the transform function (`transform_fn`)
    (typically defined by package authors) is invoked.

    The transform function (`transform_fn`) is given `meta` information
    (:class:`~shiny.render.transformer.TranformerMetadata`), the (app-supplied) value
    function (`ValueFnAsync[IT]`), and any keyword arguments supplied to the render decorator
    (`P`). For consistency, the first two parameters have been (arbitrarily) implemented
    as `_meta` and `_fn`.

    Typing
    ------
    * `IT`
        * The type returned by the app-supplied output value function (`value_fn`). This
          value should contain a `None` value to conform to the convention of app authors
          being able to return `None` to display nothing in the rendered output. Note that
          in many cases but not all, `IT` and `OT` will be the same.
    * `OT`
        * The type of the object returned by the transform function (`transform_fn`). This
          value should contain a `None` value to conform to display nothing in the
          rendered output.
    * `P`
        * The parameter specification defined by the transform function (`transform_fn`).
          It should **not** contain any `*args`. All keyword arguments should have a type
          and default value.

    See Also
    --------
    * :class:`~shiny.render.transformer.OutputRendererSync`
    * :class:`~shiny.render.transformer.OutputRendererAsync`
    """

    def __init__(
        self,
        *,
        value_fn: ValueFn[IT],
        transform_fn: TransformFn[IT, P, OT],
        params: TransformerParams[P],
        default_ui: Optional[DefaultUIFn] = None,
        default_ui_passthrough_args: Optional[tuple[str, ...]] = None,
    ) -> None:
        """
        Parameters
        ----------
        value_fn
            App-provided output value function. It should return an object of type `IT`.
        transform_fn
            Package author function that transforms an object of type `IT` into type
            `OT`. The `params` will used as variadic keyword arguments.
        params
            App-provided parameters for the transform function (`transform_fn`).
        default_ui
            Optional function that takes an `output_id` string and returns a Shiny UI
            object that can be used to display the output. This allows render functions
            to respond to `_repr_html_` method calls in environments like Jupyter.
        """
        super().__init__()

        warn_deprecated(
            "`shiny.render.transformer.output_transformer()` and `shiny.render.transformer.OutputRenderer()` output render function utilities have been superseded by `shiny.render.renderer.Renderer` and will be removed in the near future."
        )

        if not is_async_callable(transform_fn):
            raise TypeError(
                self.__class__.__name__
                + " requires an async tranformer function (`transform_fn`)."
                " Please define your transform function as asynchronous."
                " Ex `async def my_transformer(....`"
            )

        self._value_fn = value_fn
        self._value_fn_is_async = is_async_callable(value_fn)  # legacy key
        # Copy over function name as it is consistent with how Session and Output
        # retrieve function names
        self.__name__ = value_fn.__name__
        # Initial value for output_id until it is set by the Session
        self.output_id = value_fn.__name__

        self._transformer = transform_fn
        self._params = params
        self._default_ui = default_ui
        self._default_ui_passthrough_args = default_ui_passthrough_args

        self._default_ui_args: tuple[object, ...] = tuple()
        self._default_ui_kwargs: dict[str, object] = dict()

        # Register the value function with the parent Renderer class
        self(value_fn)

    def _meta(self) -> TransformerMetadata:
        """
        Returns a named tuple of values: `session` (the :class:`~shiny.Session` object),
        and `name` (the name of the output being rendered)
        """
        from ...session import require_active_session

        session = require_active_session(None)
        return TransformerMetadata(
            session=session,
            name=self.output_id,
        )

    async def _run(self) -> OT:
        """
        Executes the (async) tranform function

        The transform function will receive the following arguments: meta information of
        type :class:`~shiny.render.transformer.TransformerMetadata`, an app-defined
        render function of type :class:`~shiny.render.RenderFnAsync`, and `*args` and
        `**kwargs` of type `P`.

        Note: `*args` will always be empty as it is an expansion of
        :class:`~shiny.render.transformer.TransformerParams` which does not allow positional arguments.
        `*args` is required to use with `**kwargs` when using
        `typing.ParamSpec`.
        """
        ret = await self._transformer(  # pyright: ignore
            # TransformerMetadata
            self._meta(),
            # Callable[[], IT] | Callable[[], Awaitable[IT]]
            self._value_fn,
            # P
            *self._params.args,
            **self._params.kwargs,
        )
        return ret

    # # Shims for Renderer class #############################

    def auto_output_ui(
        self,
        **kwargs: object,
    ) -> DefaultUIFnResultOrNone:
        if self._default_ui is None:
            return None

        if self._default_ui_passthrough_args is not None:
            kwargs.update(
                {
                    k: v
                    for k, v in self._params.kwargs.items()
                    if k in self._default_ui_passthrough_args and v is not MISSING
                }
            )

        return self._default_ui(self.output_id, *self._default_ui_args, **kwargs)

    async def render(self) -> Jsonifiable:
        ret = await self._run()
        # Really, OT should be bound by Jsonifiable.
        # But we can't do that now as types like TypedDict break on Jsonifiable
        # (We also don't really care as we're moving to `Renderer` class)
        jsonifiable_ret = cast(Jsonifiable, ret)
        return jsonifiable_ret


# Using a second class to help clarify that it is of a particular type
class OutputRendererSync(OutputRenderer[OT]):
    """
    Output Renderer (Synchronous)

    See Also
    --------
    * :class:`~shiny.render.transformer.OutputRenderer`
    * :class:`~shiny.render.transformer.OutputRendererAsync`
    """

    def __init__(
        self,
        value_fn: ValueFnSync[IT],
        transform_fn: TransformFn[IT, P, OT],
        params: TransformerParams[P],
        default_ui: Optional[DefaultUIFn] = None,
        default_ui_passthrough_args: Optional[tuple[str, ...]] = None,
    ) -> None:
        if is_async_callable(value_fn):
            raise TypeError(
                self.__class__.__name__ + " requires a synchronous render function"
            )
        # super == OutputRenderer[OT]
        super().__init__(
            value_fn=value_fn,
            transform_fn=transform_fn,
            params=params,
            default_ui=default_ui,
            default_ui_passthrough_args=default_ui_passthrough_args,
        )


# The reason for having a separate RendererAsync class is because the _run
# method is marked here as async; you can't have a single class where one method could
# be either sync or async.
class OutputRendererAsync(OutputRenderer[OT]):
    """
    Output Renderer (Asynchronous)

    See Also
    --------
    * :class:`~shiny.render.transformer.OutputRenderer`
    * :class:`~shiny.render.transformer.OutputRendererSync`
    """

    def __init__(
        self,
        value_fn: ValueFnAsync[IT],
        transform_fn: TransformFn[IT, P, OT],
        params: TransformerParams[P],
        default_ui: Optional[DefaultUIFn] = None,
        default_ui_passthrough_args: Optional[tuple[str, ...]] = None,
    ) -> None:
        if not is_async_callable(value_fn):
            raise TypeError(
                self.__class__.__name__ + " requires an asynchronous render function"
            )

        # super == OutputRenderer[OT]
        super().__init__(
            value_fn=value_fn,
            transform_fn=transform_fn,
            params=params,
            default_ui=default_ui,
            default_ui_passthrough_args=default_ui_passthrough_args,
        )


# ======================================================================================
# Restrict the transformer function
# ======================================================================================


# assert: No variable length positional values;
# * We need a way to distinguish between a plain function and args supplied to the next
#   function. This is done by not allowing `*args`.
# assert: All `**kwargs` of transformer should have a default value
# * This makes calling the method with both `()` and without `()` possible / consistent.
def _assert_transformer(transform_fn: TransformFn[IT, P, OT]) -> None:
    params = inspect.Signature.from_callable(transform_fn).parameters

    if len(params) < 2:
        raise TypeError(
            "`transformer=` must have 2 positional parameters which have type "
            "`TransformerMetadata` and `ValueFnAsync` respectively"
        )

    for i, param in zip(range(len(params)), params.values()):
        # # Not a good test as `param.annotation` has type `str` and the type could
        # # have been renamed. We need to do an `isinstance` check but do not have
        # # access to the objects
        # if i == 0:
        #     assert param.annotation == "TransformerMetadata"
        # if i == 1:
        #     assert (param.annotation or "").startswith("RenderFnAsync")
        if i < 2 and not (
            param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            or param.kind == inspect.Parameter.POSITIONAL_ONLY
        ):
            raise TypeError(
                "`transformer=` must have 2 positional parameters which have type "
                "`TransformerMetadata` and `RenderFnAsync` respectively"
            )

        # Make sure there are no more than 2 positional args
        if i >= 2 and param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            raise TypeError(
                "`transformer=` must not contain more than 2 positional parameters"
            )
        # Make sure there are no `*args`
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            raise TypeError(
                "No variadic positional parameters (e.g. `*args`) can be supplied to "
                f"`transformer=`. Received: `{param.name}`. Please only use `*`."
            )
        # Make sure kwargs have default values
        if (
            param.kind == inspect.Parameter.KEYWORD_ONLY
            and param.default is inspect.Parameter.empty
        ):
            raise TypeError(
                f"In `transformer=`, parameter `{param.name}` did not have a default value"
            )


# ======================================================================================
# OutputTransformer
# ======================================================================================


# Signature of a renderer decorator function
OutputRendererDecorator = Callable[[ValueFn[IT]], OutputRenderer[OT]]
"""
Decorator function that takes the output value function (then calls it and transforms
the value) and returns an :class:`~shiny.render.transformer.OutputRenderer`.
"""

# Signature of a decorator that can be called with and without parentheses
# With parens returns a `OutputRenderer[OT]`
# Without parens returns a `OutputRendererDeco[IT, OT]`
OutputTransformerFn = Callable[
    [
        Optional[ValueFn[IT]],
        TransformerParams[P],
    ],
    Union[OutputRenderer[OT], OutputRendererDecorator[IT, OT]],
]
"""
Generic overload definition for a decorator that can be called with and without
parentheses. If called with parentheses, it returns an decorator which returns an
:class:`~shiny.render.transformer.OutputRenderer`. If called without parentheses, it
returns an :class:`~shiny.render.transformer.OutputRenderer`.
"""


class OutputTransformer(Generic[IT, OT, P]):
    """
    Output Transformer class

    This class creates helper types and methods for defining an overloaded renderer
    decorator. By manually defining the overloads locally, the function signatures are
    as clean as possible (and therefore easier to read and understand).

    When called, an `OutputTransfomer` takes the value returned from the app-supplied
    output value function and any app-supplied paramters and passes them through the
    component author's transformer function, and returns the transformed result.

    Attributes
    ----------
    ValueFn
        The function type for the app-supplied output value function. This function may
        be both synchronous or asynchronous.
    OutputRenderer
        The return type for the overload that accepts the app-supplied output value
        function and returns an object of
        :class:`~shiny.render.transformer.OutputRenderer`.
    OutputRendererDecorator
        The return type for the overload that accepts app-supplied parameters for the
        transform function. The return value is a decorator that accepts the
        app-supplied output value function and returns an object of
        :class:`~shiny.render.transformer.OutputRenderer`.

    See Also
    --------
    * :func:`~shiny.render.transformer.output_transformer`
    * :class:`~shiny.render.transformer.TransformerParams`
    * :class:`~shiny.render.transformer.OutputRenderer`
    """

    fn: OutputTransformerFn[IT, P, OT]
    ValueFn: Type[ValueFn[IT]]
    OutputRenderer: Type[OutputRenderer[OT]]
    OutputRendererDecorator: Type[OutputRendererDecorator[IT, OT]]

    def params(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> TransformerParams[P]:
        return TransformerParams(*args, **kwargs)

    def __call__(
        self,
        value_fn: ValueFn[IT] | None,
        params: TransformerParams[P] | None = None,
    ) -> OutputRenderer[OT] | OutputRendererDecorator[IT, OT]:
        if params is None:
            params = (
                self.params()
            )  # pyright: ignore[reportCallIssue] ; Missing param spec args; False positive error as we know there should be no args.
        if not isinstance(params, TransformerParams):
            raise TypeError(
                "Expected `params` to be of type `TransformerParams` but received "
                f"`{type(params)}`. Please use `.params()` to create a "
                "`TransformerParams` object."
            )
        return self._fn(value_fn, params)

    def __init__(
        self,
        fn: OutputTransformerFn[IT, P, OT],
    ) -> None:
        self._fn = fn
        self.ValueFn = ValueFn[IT]
        self.OutputRenderer = OutputRenderer[OT]

        self.OutputRendererDecorator = OutputRendererDecorator[
            IT, OT
        ]  # pyright: ignore[reportAttributeAccessIssue] ; False positive error as the types should match the class definition. Not worrying about it as this code is deprecated.


@overload
def output_transformer(
    *,
    default_ui: Optional[DefaultUIFn] = None,
    default_ui_passthrough_args: Optional[tuple[str, ...]] = None,
) -> Callable[[TransformFn[IT, P, OT]], OutputTransformer[IT, OT, P]]: ...


@overload
def output_transformer(
    transform_fn: TransformFn[IT, P, OT],
) -> OutputTransformer[IT, OT, P]: ...


@add_example()
def output_transformer(
    transform_fn: TransformFn[IT, P, OT] | None = None,
    *,
    default_ui: Optional[DefaultUIFn] = None,
    default_ui_passthrough_args: Optional[tuple[str, ...]] = None,
) -> (
    OutputTransformer[IT, OT, P]
    | Callable[[TransformFn[IT, P, OT]], OutputTransformer[IT, OT, P]]
):
    """
    Deprecated. Please use :class:`~shiny.render.renderer.Renderer` instead.

    Output transformer decorator.

    This decorator method is a convenience method to generate the appropriate types and
    internal implementation for an overloaded renderer method. This method will provide
    you with all the necessary types to define two different overloads: one for when the
    decorator is called without parentheses and another for when it is called with
    parentheses where app authors can pass in parameters to the renderer.

    Transform function
    ------------------

    The output renderer's transform function (`transform_fn`) is the key building block
    for `output_transformer`. It is a package author function that calls the app-defined
    output value function (`value_fn`) transforms the result of type `IT` into type
    `OT`.

    The transform function is supplied meta output information, the (app-supplied) value
    function, and any keyword arguments supplied to the output tranformer decorator:

    * The first parameter to the handler function has the class
      :class:`~shiny.render.transformer.TransformerMetadata` and is typically called
      `_meta`. This information gives context to the handler while trying to
      resolve the app-supplied value function (typically called `_fn`).
    * The second parameter is the app-defined output value function (e.g. `_fn`). It's
      return type (`IT`) determines what types can be returned by the app-supplied
      output value function. For example, if `_fn` has the type `ValueFnAsync[str | None]`,
      both the `str` and `None` types are allowed to be returned from the app-supplied
      output value function.
    * The remaining parameters are the keyword arguments (e.g. `alt:Optional[str] =
      None` or `**kwargs: object`) that app authors may supply to the renderer (when the
      renderer decorator is called with parentheses). Variadic positional parameters
      (e.g. `*args`) are not allowed. All keyword arguments should have a type and
      default value. No default value is needed for keyword arguments that are passed
      through (e.g. `**kwargs: Any`).

    The transform function's return type (`OT`) determines the output type of the
    :class:`~shiny.render.transformer.OutputRenderer`. Note that in many cases (but not
    all!) `IT` and `OT` will be the same. The `None` type should typically be defined in
    both `IT` and `OT`. If `IT` allows for `None` values, it (typically) signals that
    nothing should be rendered. If `OT` allows for `None` and returns a `None` value,
    shiny will not render the output.

    Notes
    -----

    * When defining the renderer decorator overloads, if you have extra parameters of
      `**kwargs: object`, you may get a type error about incompatible signatures. To fix
      this, you can use `**kwargs: Any` instead or add `_fn: None = None` as the first
      parameter in the overload containing the `**kwargs: object`.

    * The `transform_fn` should be defined as an asynchronous function but should only
      asynchronously yield (i.e. use `await` syntax) when the value function (the second
      parameter of type `ValueFn[IT]`) is awaitable. If the value function is not
      awaitable (i.e. it is a _synchronous_ function), then the execution of the
      transform function should also be synchronous.

    Parameters
    ----------
    transform_fn
        Asynchronous function used to determine the app-supplied output value function
        return type (`IT`), the transformed type (`OT`), and the keyword arguments (`P`)
        app authors can supply to the renderer decorator.
    default_ui
        Optional function that takes an `output_id` string and returns a Shiny UI object
        that can be used to display the output. This allows render functions to respond
        to `_repr_html_` method calls in environments like Jupyter.

    Returns
    -------
    :
        An :class:`~shiny.render.transformer.OutputTransformer` object that can be used to
        define two overloads for your renderer function. One overload is for when the
        renderer is called without parentheses and the other is for when the renderer is
        called with parentheses.
    """

    # If default_ui_passthrough_args was used, modify the default_ui function so it is
    # ready to mix in extra arguments from the decorator.
    def output_transformer_impl(
        transform_fn: TransformFn[IT, P, OT],
    ) -> OutputTransformer[IT, OT, P]:
        _assert_transformer(transform_fn)

        def renderer_decorator(
            value_fn: ValueFn[IT] | None,
            params: TransformerParams[P],
        ) -> OutputRenderer[OT] | OutputRendererDecorator[IT, OT]:
            def as_value_fn(
                fn: ValueFn[IT],
            ) -> OutputRenderer[OT]:
                if is_async_callable(fn):
                    return OutputRendererAsync(
                        fn,
                        transform_fn,
                        params,
                        default_ui,
                        default_ui_passthrough_args,
                    )
                else:
                    # To avoid duplicate work just for a typeguard, we cast the function
                    fn = cast(ValueFnSync[IT], fn)
                    return OutputRendererSync(
                        fn,
                        transform_fn,
                        params,
                        default_ui,
                        default_ui_passthrough_args,
                    )

            if value_fn is None:
                return as_value_fn
            val = as_value_fn(value_fn)
            return val

        return OutputTransformer(renderer_decorator)

    if transform_fn is not None:
        return output_transformer_impl(transform_fn)
    else:
        return output_transformer_impl


async def resolve_value_fn(value_fn: ValueFn[IT]) -> IT:
    """
    Resolve the value function

    This function is used to resolve the value function (`value_fn`) to an object of
    type `IT`. If the value function is asynchronous, it will be awaited. If the value
    function is synchronous, it will be called.

    While always using an async method within an output transform function is not
    appropriate, this method may be safely used to avoid boilerplate.

    Replace this:
    ```python
    if is_async_callable(_fn):
        x = await _fn()
    else:
        x = cast(ValueFnSync[IT], _fn)()
    ```

    With this:
    ```python
    x = await resolve_value_fn(_fn)
    ```

    Parameters
    ----------
    value_fn
        App-supplied output value function which returns type `IT`. This function can be
        synchronous or asynchronous.

    Returns
    -------
    :
        The resolved value from `value_fn`.
    """
    if is_async_callable(value_fn):
        return await value_fn()
    else:
        # To avoid duplicate work just for a typeguard, we cast the function
        value_fn = cast(ValueFnSync[IT], value_fn)
        return value_fn()



from __future__ import annotations

import great_tables as gt

from shiny.render._data_frame_utils._styles import StyleInfo
from shiny.types import Jsonifiable


def gt_style_str_to_obj(style_str: str) -> dict[str, Jsonifiable]:
    # Could use BeautifulSoup here, but this is a simple example
    style_obj: dict[str, Jsonifiable] = {}
    for style_part in style_str.split(";"):
        style_part = style_part.strip()
        if not style_part:
            continue
        key, value = style_part.split(":")
        style_obj[key] = value.strip()
    return style_obj


def gt_styles(df_gt: gt.GT) -> list[StyleInfo]:
    styles = df_gt._styles
    ret: list[StyleInfo] = []
    for style in styles:
        location = style.locname
        location = (
            "body"
            if location == "data"  # pyright: ignore[reportUnnecessaryComparison]
            else location
        )
        assert location == "body", f"`style.locname` is {location}, expected 'body'"
        rows = style.rownum
        assert rows is not None
        cols = style.colnum
        if cols is None:
            cols = style.colname
        assert cols is not None

        style_obj: dict[str, Jsonifiable] = {}
        for style_str in style.styles:
            style_obj.update(
                gt_style_str_to_obj(style_str._to_html_style()),
            )
        ret.append(
            {
                "location": location,  # pyright: ignore[reportArgumentType]
                "rows": rows,
                "cols": cols,
                "style": style_obj,
            }
        )
    return ret



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("render_express")

EXPECT_TIMEOUT = 30 * 1000


def test_render_express(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)
    expect(page.get_by_text("Name")).to_have_count(1, timeout=EXPECT_TIMEOUT)
    expect(page.get_by_text("Socrates")).to_have_count(1, timeout=EXPECT_TIMEOUT)



import matplotlib.pyplot as plt
import numpy as np

from shiny import render
from shiny.express import input, ui

with ui.layout_column_wrap(width=1 / 2):
    with ui.card():
        ui.input_slider("n", "N", 1, 100, 50)

    with ui.card():

        @render.plot
        def histogram():
            np.random.seed(19680801)
            x = 100 + 15 * np.random.randn(437)
            plt.hist(x, input.n(), density=True)

    with ui.card():

        @render.plot
        def histogram2():
            np.random.seed(19680801)
            x = 100 + 15 * np.random.randn(437)
            plt.hist(x, input.n(), density=True, color="red")



import pytest
from playwright.sync_api import Page
from utils.deploy_utils import (
    local_deploys_app_url_fixture,
    reruns,
    reruns_delay,
    skip_if_not_chrome,
)

from shiny.playwright import controller

app_url = local_deploys_app_url_fixture("shiny-express-dataframe")


@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_express_dataframe_deploys(page: Page, app_url: str) -> None:
    page.goto(app_url)

    dataframe = controller.OutputDataFrame(page, "sample_data_frame")
    dataframe.expect_nrow(6)



from shared import mtcars

from shiny import App, reactive, render, ui

app_ui = ui.page_fillable(
    ui.card(
        ui.layout_column_wrap(
            ui.input_action_button("btn", "Sort on columns 1↑ and 3↓"),
            ui.input_action_button("reset", "Reset sorting"),
            fill=False,
        ),
        ui.output_data_frame("df"),
    ),
)


def server(input, output, session):
    data = reactive.value(mtcars.iloc[:, range(4)])

    @render.data_frame
    def df():
        return render.DataGrid(data())

    @reactive.effect
    @reactive.event(input.reset)
    async def _():
        await df.update_sort(None)

    @reactive.effect
    @reactive.event(input.btn)
    async def _():
        await df.update_sort([{"col": 1, "desc": False}, {"col": 3, "desc": True}])


app = App(app_ui, server, debug=True)



from __future__ import annotations

import platform

from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.render._data_frame_utils._types import ColumnFilterStr
from shiny.run import ShinyAppProc


def test_row_selection(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    my_df = controller.OutputDataFrame(page, "my_df")
    selected_df = controller.OutputDataFrame(page, "selected_df")
    selected_rows = controller.OutputCode(page, "selected_rows")

    # TODO-karan-test: We should add a locator for selected rows
    # TODO-karan-test; We should add a expected selected row count method?
    def expect_selected_count(x: controller.OutputDataFrame, n: int) -> None:
        expect(x.loc_body.locator("tr[aria-selected=true]")).to_have_count(n)

    my_df.expect_nrow(10)
    expect_selected_count(my_df, 0)
    selected_df.expect_nrow(0)
    selected_rows.expect_value("()")

    # Select row
    my_df.select_rows([5, 7])
    my_df.expect_nrow(10)
    expect_selected_count(my_df, 2)
    selected_df.expect_nrow(2)
    selected_rows.expect_value("(5, 7)")

    filter_text: ColumnFilterStr = {"col": 0, "value": "5"}
    # # Filter to different row
    # Currently this causes an error
    my_df.set_filter(filter_text)
    my_df.expect_nrow(1)
    expect_selected_count(my_df, 1)
    selected_df.expect_nrow(1)
    selected_rows.expect_value("(5,)")
    # # Remove the filter
    my_df.set_filter(None)
    # Confirm the original data frame returns
    my_df.expect_nrow(10)
    # Confirm the previous row is still selected as no new selection has been made
    expect_selected_count(my_df, 2)
    selected_df.expect_nrow(2)
    selected_rows.expect_value("(5, 7)")

    # Filter to non selected row
    # Currently this causes an error
    filter_text["value"] = "8"
    filter_text["col"] = 0
    my_df.set_filter(filter_text)
    my_df.expect_nrow(1)
    expect_selected_count(my_df, 0)
    selected_df.expect_nrow(0)
    selected_rows.expect_value("()")

    # Remove the filter
    # Confirm the original data frame returns
    my_df.set_filter(None)
    my_df.expect_nrow(10)
    expect_selected_count(my_df, 2)
    selected_df.expect_nrow(2)
    selected_rows.expect_value("(5, 7)")

    # Update selection with new selection while under a filter
    my_df.set_filter(filter_text)
    my_df.expect_nrow(1)
    expect_selected_count(my_df, 0)
    selected_df.expect_nrow(0)
    selected_rows.expect_value("()")
    my_df.select_rows([0])
    my_df.set_filter(None)
    my_df.expect_nrow(10)
    expect_selected_count(my_df, 1)
    selected_df.expect_nrow(1)
    selected_rows.expect_value("(8,)")

    # Remove selection
    modifier = "Meta" if platform.system() == "Darwin" else "Control"
    for row in (8,):
        my_df.cell_locator(row=row, col=0).click(
            modifiers=[modifier]
        )  # TODO-karan-test: Support unselect row method
    expect_selected_count(my_df, 0)
    selected_df.expect_nrow(0)
    selected_rows.expect_value("()")



import htmltools

from shiny import App, ui
from shiny._app import html_dep_name


def test_duplicate_deps():
    some_dep1 = htmltools.HTMLDependency(
        "test_dep",
        "1",
        source={
            "subdir": ".",
        },
        all_files=True,
    )
    some_dep2 = htmltools.HTMLDependency(
        "test_dep",
        "2",
        source={
            "subdir": ".",
        },
        all_files=True,
    )
    some_dep3 = htmltools.HTMLDependency(
        "test_dep",
        "3",
        source={
            "subdir": ".",
        },
        all_files=True,
    )

    # Put v1 at the end to make sure lower versions can be added
    # Put v3 in the middle to make sure a higher version can be added
    deps = [some_dep2, some_dep3, some_dep1]

    app = App(ui.div("ui goes here"), server=None)

    # During a page refresh, the same session is kept alive. This means the mapping of `app._registered_dependencies` is kept. However, the dependencies requested by the user can be different when using a UI function.

    # Simulate adding the _conflicting_ dependencies from three different dynamic UI
    # functions
    for dep in deps:
        app._register_web_dependency(dep)
    # All three dependencies should be registered as they are unique and will be requested by the browser
    for dep in deps:
        assert html_dep_name(dep) in app._registered_dependencies



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_numeric_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default = controller.InputNumeric(page, "default")
    default.expect_label("Default numeric input")
    default.expect_value("10")
    controller.OutputUi(page, "default_txt").expect.to_have_text("10")

    min_max = controller.InputNumeric(page, "min_max")
    min_max.expect_max("100")
    min_max.expect_min("0")
    min_max.expect_value("50")
    controller.OutputUi(page, "min_max_txt").expect.to_have_text("50")

    step = controller.InputNumeric(page, "step")
    step.expect_step("0.5")

    width = controller.InputNumeric(page, "width")
    width.expect_width("200px")
    width.set("20")
    width.expect_value("20")
    controller.OutputUi(page, "width_txt").expect.to_have_text("20")



import os
import subprocess
import tempfile

import pytest
from example_apps import get_apps, reruns, reruns_delay, validate_example
from playwright.sync_api import Page

from shiny._main_create import (
    GithubRepoLocation,
    parse_github_arg,
    shiny_internal_templates,
)


def subprocess_create(
    app_template: str,
    dest_dir: str = "",
    *,
    mode: str = "core",
    package_name: str = "",
):
    cmd = [
        "shiny",
        "create",
        "--template",
        app_template,
        "--mode",
        mode,
        "--dir",
        dest_dir,
        "--package-name",
        package_name,
    ]

    subprocess.run(cmd)
    assert os.path.isdir(dest_dir)

    # Package templates don't have a top-level app.py file
    if package_name == "":
        assert os.path.isfile(f"{dest_dir}/app.py")
    else:
        assert os.path.isfile(f"{dest_dir}/pyproject.toml")

    print("\nCheck for duplicate files")
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.run(cmd, check=True)


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("ex_app_path", get_apps("shiny/templates/app"))
def test_template_examples(page: Page, ex_app_path: str) -> None:
    validate_example(page, ex_app_path)


app_templates = [t.id for t in shiny_internal_templates.apps]
pkg_templates = [t.id for t in shiny_internal_templates.packages]
assert len(app_templates) > 0
assert len(pkg_templates) > 0


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("app_template", app_templates)
def test_create_core(app_template: str, page: Page):
    with tempfile.TemporaryDirectory("example_apps") as tmpdir:
        subprocess_create(app_template, dest_dir=tmpdir)
        validate_example(page, f"{tmpdir}/app.py")


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("app_template", ["basic-app"])
def test_create_express(app_template: str, page: Page):
    with tempfile.TemporaryDirectory("example_apps") as tmpdir:
        subprocess_create(app_template, dest_dir=tmpdir, mode="express")
        validate_example(page, f"{tmpdir}/app.py")


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("app_template", pkg_templates)
def test_create_js(app_template: str):
    with tempfile.TemporaryDirectory("example_apps") as tmpdir:
        subprocess_create(app_template, dest_dir=tmpdir, package_name="my_component")
        # TODO-Karan: Add test to validate once flag to install packages is implemented


def test_parse_github_arg():
    expected = GithubRepoLocation(
        repo_owner="posit-dev",
        repo_name="py-shiny",
        ref="main",
        path="shiny/templates/app/basic-app",
    )

    # * {repo_owner}/{repo_name}@{ref}:{path}
    actual_ref_path = parse_github_arg(
        "posit-dev/py-shiny@main:shiny/templates/app/basic-app"
    )
    assert actual_ref_path == expected

    # * {repo_owner}/{repo_name}:{path}@{ref}
    actual_path_ref = parse_github_arg(
        "posit-dev/py-shiny:shiny/templates/app/basic-app@main"
    )
    assert actual_path_ref == expected

    # * {repo_owner}/{repo_name}/{path}@{ref}
    actual_path_slash_ref = parse_github_arg(
        "posit-dev/py-shiny/shiny/templates/app/basic-app@main"
    )
    assert actual_path_slash_ref == expected

    # * {repo_owner}/{repo_name}/{path}?ref={ref}
    actual_path_slash_query = parse_github_arg(
        "posit-dev/py-shiny/shiny/templates/app/basic-app?ref=main"
    )
    assert actual_path_slash_query == expected

    actual_path_full = parse_github_arg(
        "https://github.com/posit-dev/py-shiny/tree/main/shiny/templates/app/basic-app"
    )
    assert actual_path_full == expected

    actual_path_part = parse_github_arg(
        "github.com/posit-dev/py-shiny/tree/main/shiny/templates/app/basic-app"
    )
    assert actual_path_part == expected

    # REF is implied, defaults to HEAD
    expected.ref = "HEAD"

    # * {repo_owner}/{repo_name}:{path}
    actual_path_colon = parse_github_arg(
        "posit-dev/py-shiny:shiny/templates/app/basic-app"
    )
    assert actual_path_colon == expected

    # * {repo_owner}/{repo_name}/{path}
    actual_path_slash = parse_github_arg(
        "posit-dev/py-shiny/shiny/templates/app/basic-app"
    )
    assert actual_path_slash == expected

    # complicated ref
    actual_ref_tag = parse_github_arg(
        "posit-dev/py-shiny@v0.1.0:shiny/templates/app/basic-app"
    )
    expected.ref = "v0.1.0"
    assert actual_ref_tag == expected

    actual_ref_branch = parse_github_arg(
        "posit-dev/py-shiny@feat/new-template:shiny/templates/app/basic-app"
    )
    expected.ref = "feat/new-template"
    assert actual_ref_branch == expected



from __future__ import annotations

import time
import typing
from typing import Literal

from playwright.sync_api import FloatRect, Locator, Page, Position
from playwright.sync_api import expect as playwright_expect

from ...types import MISSING, MISSING_TYPE, ListOrTuple
from .._types import AttrValue, ListPatternOrStr, PatternOrStr, Timeout
from ..expect._expect import _attr_match_str
from ..expect._internal import (
    expect_attribute_to_have_value as _expect_attribute_to_have_value,
)
from ..expect._internal import expect_class_to_have_value as _expect_class_to_have_value
from ..expect._internal import expect_style_to_have_value as _expect_style_to_have_value
from ._base import (
    InitLocator,
    UiWithLabel,
    WidthContainerStyleM,
    all_missing,
    not_is_missing,
)
from ._expect import (
    expect_locator_contains_values_in_list,
    expect_locator_values_in_list,
)


class _InputSliderBase(
    WidthContainerStyleM,
    UiWithLabel,
):

    loc_irs: Locator
    """
    Playwright `Locator` of the input slider.
    """
    loc_irs_ticks: Locator
    """
    Playwright `Locator` of the input slider ticks.
    """
    loc_play_pause: Locator
    """
    Playwright `Locator` of the play/pause button.
    """

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initialize the InputSlider.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The id of the slider.
        """
        super().__init__(
            page,
            id=id,
            loc=f"input#{id}",
        )
        self.loc_irs = self.loc_container.locator("> .irs.irs--shiny")
        self.loc_irs_ticks = self.loc_irs.locator("> .irs-grid > .irs-grid-text")
        self.loc_play_pause = self.loc_container.locator(
            "> .slider-animate-container a"
        )

    def expect_tick_labels(
        self,
        value: ListPatternOrStr | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the tick labels of the input slider.

        Parameters
        ----------
        value
            The expected tick labels.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        if value is None:
            playwright_expect(self.loc_irs_ticks).to_have_count(0)
            return

        playwright_expect(self.loc_irs_ticks).to_have_text(value, timeout=timeout)

    def expect_animate(self, exists: bool, *, timeout: Timeout = None) -> None:
        """
        Expect the animate button to exist.

        Parameters
        ----------
        exists
            Whether the animate button should exist.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        animate_count = 1 if exists else 0
        playwright_expect(self.loc_play_pause).to_have_count(animate_count)

    # This method doesn't feel like it should accept text as the user does not control the value
    # They only control either `True` or `False`
    def expect_animate_options(
        self,
        *,
        loop: bool | MISSING_TYPE = MISSING,
        interval: float | MISSING_TYPE = MISSING,
        timeout: Timeout = None,
    ) -> None:
        if all_missing(loop, interval):
            raise ValueError("Must provide at least one of `loop` or `interval`")
        # TODO-future; Composable expectations
        self.expect_animate(exists=True, timeout=timeout)
        if not_is_missing(loop):
            _expect_attribute_to_have_value(
                self.loc_play_pause,
                "data-loop",
                "" if loop else None,
                timeout=timeout,
            )
        if not_is_missing(interval):
            _expect_attribute_to_have_value(
                self.loc_play_pause,
                "data-interval",
                str(interval),
                timeout=timeout,
            )

    # No `toggle` method as short animations with no loops can cause the button to
    # become `play` over and over again. Instead, have explicit `play` and `pause`
    # methods.
    def click_play(self, *, timeout: Timeout = None) -> None:
        """
        Click the play button.

        Parameters
        ----------
        timeout
            The timeout for the action. Defaults to `None`.
        """
        self.loc_container.wait_for(state="visible", timeout=timeout)
        self.loc_container.scroll_into_view_if_needed(timeout=timeout)
        _expect_class_to_have_value(
            self.loc_play_pause,
            "playing",
            has_class=False,
            timeout=timeout,
        )
        self.loc_play_pause.click()

    def click_pause(self, *, timeout: Timeout = None) -> None:
        """
        Click the pause button.

        Parameters
        ----------
        timeout
            The timeout for the action. Defaults to `None`.
        """
        self.loc_container.wait_for(state="visible", timeout=timeout)
        self.loc_container.scroll_into_view_if_needed(timeout=timeout)
        _expect_class_to_have_value(
            self.loc_play_pause, "playing", has_class=True, timeout=timeout
        )
        self.loc_play_pause.click()

    def expect_min(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the input element to have the expected `min` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-min", value=value, timeout=timeout
        )

    def expect_max(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the input element to have the expected `max` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-max", value=value, timeout=timeout
        )

    def expect_step(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the input element to have the expected `step` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-step", value=value, timeout=timeout
        )

    def expect_ticks(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the input element to have the expected `data-ticks` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-grid", value=value, timeout=timeout
        )

    def expect_sep(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the input element to have the expected `data-sep` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-prettify-separator", value=value, timeout=timeout
        )

    def expect_pre(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the input element to have the expected `data-pre` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-prefix", value=value, timeout=timeout
        )

    def expect_post(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the input element to have the expected `data-post` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-postfix", value=value, timeout=timeout
        )

    # def expect_data_type(
    #     self, value: AttrValue, *, timeout: Timeout = None
    # ) -> None:
    #     expect_attr(self.loc, "data-data-type", value=value, timeout=timeout)

    def expect_time_format(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Asserts that the input element has the expected `data-time-format` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-time-format", value=value, timeout=timeout
        )

    def expect_timezone(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Asserts that the input element has the expected `data-timezone` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-timezone", value=value, timeout=timeout
        )

    def expect_drag_range(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Asserts that the input element has the expected `data-drag-range` attribute value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-drag-interval", value=value, timeout=timeout
        )

    def _wait_for_container(self, *, timeout: Timeout = None) -> None:
        self.loc_container.wait_for(state="visible", timeout=timeout)
        self.loc_container.scroll_into_view_if_needed(timeout=timeout)

    def _set_helper(
        self,
        *,
        value: str,
        irs_label: Locator,
        handle_center: Position,
        grid_bb: FloatRect,
        start_x: float,
        direction: Literal["left", "right"],
        max_err_values: int = 15,
    ) -> None:
        if direction == "left":
            pixel_increment = -1
            error_msg_direction = "right to left"
            min_pxls = -1 * (grid_bb["width"] + 1)

            def should_continue(cur_pxls: float) -> bool:
                return cur_pxls > min_pxls

        elif direction == "right":
            pixel_increment = 1
            error_msg_direction = "left to right"
            max_pxls = grid_bb["width"] + 1

            def should_continue(cur_pxls: float) -> bool:
                return cur_pxls <= max_pxls

        else:
            raise ValueError(f"Invalid direction: {direction}")

        # Move mouse to handle center and press down on mouse
        mouse = self.loc_container.page.mouse
        mouse.move(handle_center.get("x"), handle_center.get("y"))
        mouse.down()

        # Slow it down a bit. Like "type" for text, but to allow the slider label to catch up
        # This should be done for every `mouse.move()` call
        sleep_time = 0.05

        def slow_move(x: float, y: float, delay: float = sleep_time) -> None:
            """
            Slowly move the mouse to the given coordinates.

            Parameters
            ----------
            x
                The x-coordinate.
            y
                The y-coordinate.
            delay
                The delay between each move. Defaults to `sleep_time`.
            """
            mouse.move(x, y)
            time.sleep(delay)

        # Move all the way to the left
        handle_center_y = handle_center.get("y")
        slow_move(start_x, handle_center_y, delay=10 * sleep_time)

        # For each pixel in the grid width, check the text label
        pxls: int = 0
        found = False
        values_found: typing.Dict[str, bool] = {}
        cur_val = None
        while should_continue(pxls):
            # Get value
            cur_val = irs_label.inner_text()
            # Only store what could be used
            if len(values_found) <= max_err_values + 1:
                values_found[cur_val] = True

            # Quit if found
            if cur_val == value:
                found = True
                break

            # Not found; move handle to the right
            slow_move(start_x + pxls, handle_center_y)

            pxls += pixel_increment

        mouse.up()
        if not found:
            key_arr = list(values_found.keys())
            trail_txt = ""
            if len(key_arr) > max_err_values:
                key_arr = key_arr[:max_err_values]
                trail_txt = f", ...\nTo display more values, increase `set(max_err_values={max_err_values})`"
            values_found_txt = ", ".join([f'"{key}"' for key in key_arr])
            raise ValueError(
                f"Could not find value '{value}' when moving slider from {error_msg_direction}\n"
                + f"Values found:\n{values_found_txt}{trail_txt}"
            )

    def _grid_bb(self, *, timeout: Timeout = None) -> FloatRect:
        grid = self.loc_irs.locator("> .irs > .irs-line")
        grid_bb = grid.bounding_box(timeout=timeout)
        if grid_bb is None:
            raise RuntimeError("Couldn't find bounding box for .irs-line")
        return grid_bb

    def _handle_center(
        self,
        handle: Locator,
        *,
        name: str,
        timeout: Timeout = None,
    ) -> Position:
        handle_bb = handle.bounding_box(timeout=timeout)
        if handle_bb is None:
            raise RuntimeError(f"Couldn't find bounding box for {name}")

        handle_center: Position = {
            "x": handle_bb.get("x") + (handle_bb.get("width") / 2),
            "y": handle_bb.get("y") + (handle_bb.get("height") / 2),
        }
        return handle_center


class _RadioButtonCheckboxGroupBase(
    WidthContainerStyleM,
    UiWithLabel,
):
    loc_choice_labels: Locator

    def expect_choice_labels(
        self,
        value: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the labels of the choices.

        Parameters
        ----------
        value
            The expected labels.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        if len(value) == 1:
            labels_val = value[0]
        else:
            labels_val = value
        playwright_expect(self.loc_choice_labels).to_have_text(
            labels_val,
            timeout=timeout,
        )

    def expect_inline(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expect the input to be inline.

        Parameters
        ----------
        value
            Whether the input is inline.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc_container,
            "shiny-input-container-inline",
            has_class=value,
            timeout=timeout,
        )


class InputRadioButtons(
    _RadioButtonCheckboxGroupBase,
):
    """Controller for :func:`shiny.ui.input_radio_buttons`."""

    loc_selected: Locator
    """
    Playwright `Locator` of the selected radio button.
    """
    loc_choices: Locator
    """
    Playwright `Locator` of the radio button choices.
    """
    loc_choice_labels: Locator
    """
    Playwright `Locator` of the labels of the radio button choices.
    """

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """Initialize the InputRadioButtons.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The id of the radio buttons.
        """
        super().__init__(
            page,
            id=id,
            # Similar to `select` tag in `InputSelect`'s `loc`
            # loc should be the `.shiny-bound-input` element
            # This happens to be the container
            loc="xpath=.",
            loc_container=f"div#{id}.shiny-input-radiogroup.shiny-bound-input",
        )

        # # Regular example
        #     <div class="shiny-options-group">
        #       <div class="radio">
        #         <label>
        #           <input type="radio" name="radio1" value="a" checked="checked">
        #           <span><span style="color:red;">A</span></span>
        # # Inline example
        #     <div class="shiny-options-group">
        #       <label class="radio-inline">
        #         <input type="radio" name="radio2" value="d" checked="checked">
        #         <span><span style="color:purple;">D</span></span>
        input_radio = f"> .shiny-options-group input[type=radio][name={id}]"
        self.loc_selected = self.loc.locator(f"{input_radio}:checked")
        self.loc_choices = self.loc.locator(f"{input_radio}")
        # Get sibling <span> containing the label text
        self.loc_choice_labels = self.loc.locator(f"{input_radio} + span")

    def set(
        self,
        selected: str,
        *,
        timeout: Timeout = None,
        **kwargs: object,
    ) -> None:
        """
        Set the selected radio button.

        Parameters
        ----------
        selected
            The value of the selected radio button.
        timeout
            The timeout for the action. Defaults to `None`.
        """
        if not isinstance(selected, str):
            raise TypeError("`selected` must be a string")

        # Only need to set.
        # The Browser will _unset_ the previously selected radio button
        self.loc_container.locator(
            f"label input[type=radio][{_attr_match_str('value', selected)}]"
        ).check(timeout=timeout)

    def expect_choices(
        self,
        value: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the radio button choices.

        Parameters
        ----------
        value
            The expected choices.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type="input[type=radio]",
            arr_name="choices",
            arr=value,
            timeout=timeout,
        )

    def expect_selected(
        self,
        value: PatternOrStr | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the selected radio button.

        Parameters
        ----------
        value
            The expected value of the selected radio button.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        # Playwright doesn't like lists of size 0. Instead, use `None`
        if value is None:
            playwright_expect(self.loc_selected).to_have_count(0, timeout=timeout)
            return

        playwright_expect(self.loc_selected).to_have_value(value, timeout=timeout)


class _InputCheckboxBase(
    WidthContainerStyleM,
    UiWithLabel,
):
    def __init__(
        self, page: Page, id: str, loc: InitLocator, loc_label: str | None
    ) -> None:
        """
        Initializes the input checkbox.

        Parameters
        ----------
        page
            The page where the input checkbox is located.
        id
            The id of the input checkbox.
        loc
            Playwright `Locator` of the input checkbox.
        loc_label
            Playwright `Locator` of the label of the input checkbox.
        """
        super().__init__(
            page,
            id=id,
            loc=loc,
            loc_label=loc_label,
        )

    def set(self, value: bool, *, timeout: Timeout = None, **kwargs: object) -> None:
        """
        Sets the input checkbox.

        Parameters
        ----------
        value
            The value of the input checkbox.
        timeout
            The maximum time to wait for the input checkbox to be set. Defaults to `None`.
        """
        self.loc.wait_for(state="visible", timeout=timeout)
        self.loc.scroll_into_view_if_needed(timeout=timeout)
        self.loc.set_checked(
            value, timeout=timeout, **kwargs  # pyright: ignore[reportArgumentType]
        )

    def _toggle(self, *, timeout: Timeout = None, **kwargs: object) -> None:
        """
        Toggles the input checkbox.

        Parameters
        ----------
        timeout
            The maximum time to wait for the input checkbox to be toggled. Defaults to `None`.
        """
        self.loc.wait_for(state="visible", timeout=timeout)
        self.loc.scroll_into_view_if_needed(timeout=timeout)
        self.loc.click(timeout=timeout, **kwargs)  # pyright: ignore[reportArgumentType]

    def expect_checked(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expect the input checkbox to be checked.

        Parameters
        ----------
        value
            Whether the input checkbox is checked.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        if value:
            self.expect.to_be_checked(timeout=timeout)
        else:
            self.expect.not_to_be_checked(timeout=timeout)


class InputCheckboxGroup(
    _RadioButtonCheckboxGroupBase,
):
    """Controller for :func:`shiny.ui.input_checkbox_group`."""

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initialize the InputCheckboxGroup.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The id of the checkbox group.
        """
        super().__init__(
            page,
            id=id,
            # Similar to `select` tag in `InputSelect`'s `loc`
            # loc should be the `.shiny-bound-input` element
            # This happens to be the container
            loc="xpath=.",
            loc_container=f"div#{id}.shiny-input-checkboxgroup.shiny-bound-input",
        )

        # # Regular example
        #     <div class="shiny-options-group">
        #       <div class="checkbox">
        #         <label>
        #           <input type="checkbox" name="check1" value="red">
        #           <span><span style="color: #FF0000;">RED</span></span>
        # # Inline example
        #     <div class="shiny-options-group">
        #       <label class="checkbox-inline">
        #         <input type="checkbox" name="check2" value="magenta">
        #         <span><span style="color: #FF00AA;">MAGENTA</span></span>
        input_checkbox = f"> .shiny-options-group input[type=checkbox][name={id}]"
        self.loc_selected = self.loc.locator(f"{input_checkbox}:checked")
        self.loc_choices = self.loc.locator(f"{input_checkbox}")
        # Get sibling <span> containing the label text
        self.loc_choice_labels = self.loc.locator(f"{input_checkbox} + span")

    def set(
        self,
        # TODO-future: Allow `selected` to be a single Pattern to perform matching against many items
        selected: ListOrTuple[str],
        *,
        timeout: Timeout = None,
        **kwargs: object,
    ) -> None:
        """
        Set the selected checkboxes.

        Parameters
        ----------
        selected
            The values of the selected checkboxes.
        timeout
            The timeout for the action. Defaults to `None`.
        """
        # Having an arr of size 0 is allowed. Will uncheck everything
        if not isinstance(selected, list):
            raise TypeError("`selected` must be a list or tuple")
        for item in selected:
            if not isinstance(item, str):
                raise TypeError("`selected` must be a list of strings")

        # Make sure the selected items exist
        # Similar to `self.expect_choices(choices = selected)`, but with
        # `is_exact=False` to allow for values not in `selected`.
        expect_locator_contains_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type="input[type=checkbox]",
            arr_name="selected",
            arr=selected,
            timeout=timeout,
        )

        def in_selected(value: str) -> bool:
            for item in selected:
                if isinstance(item, str):
                    if item == value:
                        return True
                elif isinstance(item, typing.Pattern):
                    if item.search(value):
                        return True
            return False

        for i in range(self.loc_choices.count()):
            checkbox = self.loc_choices.nth(i)
            is_selected = in_selected(checkbox.input_value(timeout=timeout))
            currently_selected = checkbox.is_checked()
            # Only update if needed
            if is_selected != currently_selected:
                checkbox.set_checked(
                    is_selected,
                    timeout=timeout,
                    **kwargs,  # pyright: ignore[reportArgumentType]
                )

    def expect_choices(
        self,
        value: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the checkbox choices.

        Parameters
        ----------
        value
            The expected choices.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type="input[type=checkbox]",
            arr_name="choices",
            arr=value,
            timeout=timeout,
        )

    def expect_selected(
        self,
        value: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the selected checkboxes.

        Parameters
        ----------
        value
            The expected values of the selected checkboxes.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        # Playwright doesn't like lists of size 0
        if len(value) == 0:
            playwright_expect(self.loc_selected).to_have_count(0, timeout=timeout)
            return

        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type="input[type=checkbox]",
            arr_name="selected",
            arr=value,
            timeout=timeout,
            is_checked=True,
        )


class InputCheckbox(_InputCheckboxBase):
    """Controller for :func:`shiny.ui.input_checkbox`."""

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initializes the input checkbox.

        Parameters
        ----------
        page
            The page where the input checkbox is located.
        id
            The id of the input checkbox.
        """
        super().__init__(
            page,
            id=id,
            loc=f"div.checkbox > label > input#{id}[type=checkbox].shiny-bound-input",
            loc_label="label > span",
        )


class InputSwitch(_InputCheckboxBase):
    """Controller for :func:`shiny.ui.input_switch`."""

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initializes the input switch.

        Parameters
        ----------
        page
            The page where the input switch is located.
        id
            The id of the input switch.
        """
        super().__init__(
            page,
            id=id,
            loc=f"div.form-switch > input#{id}[type=checkbox].shiny-bound-input",
            loc_label=f"label[for={id}]",
        )


class InputSelect(
    WidthContainerStyleM,
    UiWithLabel,
):
    """
    Controller for :func:`shiny.ui.input_select`.

    If you have defined your app's select input (`ui.input_select()`) with `selectize=TRUE`, use `InputSelectize` to test your app's UI.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes the input select.

        Parameters
        ----------
        page
            The page where the input select is located.
        id
            The id of the input select.
        """
        super().__init__(
            page,
            id=id,
            loc=f"select#{id}.shiny-bound-input.form-select",
        )
        self.loc_selected = self.loc.locator("option:checked")
        self.loc_choices = self.loc.locator("option")
        self.loc_choice_groups = self.loc.locator("optgroup")

    def set(
        self,
        selected: str | ListOrTuple[str],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Sets the selected option(s) of the input select.

        Parameters
        ----------
        selected
            The value(s) of the selected option(s).
        timeout
            The maximum time to wait for the selection to be set. Defaults to `None`.
        """
        if isinstance(selected, str):
            selected = [selected]
        self.loc.select_option(value=selected, timeout=timeout)

    # If `selectize=` parameter does not become deprecated, uncomment this
    # # selectize: bool = False,
    # def expect_selectize(self, value: bool, *, timeout: Timeout = None) -> None:
    #     """
    #     Expect the input select to be selectize.

    #     Parameters
    #     ----------
    #     value
    #         Whether the input select is selectize.
    #     timeout
    #         The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
    #     """
    #     # class_=None if selectize else "form-select",
    #     _expect_class_to_have_value(
    #         self.loc,
    #         "form-select",
    #         has_class=not value,
    #         timeout=timeout,
    #     )

    def expect_choices(
        self,
        choices: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the available options of the input select to be an exact match.

        Parameters
        ----------
        choices
            The expected choices of the input select.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """

        # Playwright doesn't like lists of size 0. Instead, check for empty locator
        if len(choices) == 0:
            playwright_expect(self.loc_choices).to_have_count(0, timeout=timeout)
            return

        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type="option",
            arr_name="choices",
            arr=choices,
            timeout=timeout,
        )

    def expect_selected(
        self,
        value: PatternOrStr | ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the selected option(s) of the input select to be an exact match.

        Parameters
        ----------
        value
            The expected value(s) of the selected option(s).
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        # Playwright doesn't like lists of size 0
        if isinstance(value, list) and len(value) == 0:
            playwright_expect(self.loc_selected).to_have_count(0, timeout=timeout)
            return

        if isinstance(value, list):
            self.expect.to_have_values(value, timeout=timeout)
        else:
            self.expect.to_have_value(value, timeout=timeout)

        # expect_locator_values_in_list(
        #     page=self.page,
        #     loc_container=self.loc_container,
        #     el_type="option",
        #     arr_name="selected",
        #     arr=selected,
        #     timeout=timeout,
        #     is_checked=True,
        # )

    def expect_choice_groups(
        self,
        # TODO-future; support patterns?
        choice_groups: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the choice groups of the input select to be an exact match.

        Parameters
        ----------
        choice_groups
            The expected choice groups of the input select.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """

        # Playwright doesn't like lists of size 0. Instead, use `None`
        if len(choice_groups) == 0:
            playwright_expect(self.loc_choice_groups).to_have_count(0, timeout=timeout)
            return

        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type="optgroup",
            arr_name="choice_groups",
            arr=choice_groups,
            timeout=timeout,
            key="label",
        )

    def expect_choice_labels(
        self,
        value: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the choice labels of the input select to be an exact match.

        Parameters
        ----------
        value
            The expected choice labels of the input select.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        # Playwright doesn't like lists of size 0. Instead, use `None`
        if len(value) == 0:
            playwright_expect(self.loc_choices).to_have_count(0, timeout=timeout)
            return
        playwright_expect(self.loc_choices).to_have_text(value, timeout=timeout)

    def expect_multiple(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expect the input selectize to allow multiple selections.

        Parameters
        ----------
        value
            Whether the input select allows multiple selections.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc,
            "multiple",
            value="" if value else None,
            timeout=timeout,
        )

    def expect_size(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the size attribute of the input select to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `size` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc,
            "size",
            value=value,
            timeout=timeout,
        )


class InputSelectize(
    WidthContainerStyleM,
    UiWithLabel,
):
    """Controller for :func:`shiny.ui.input_selectize`."""

    def __init__(self, page: Page, id: str) -> None:
        super().__init__(page, id=id, loc=f"#{id} + .selectize-control")
        self._loc_dropdown = self.loc.locator("> .selectize-dropdown")
        self._loc_events = self.loc.locator("> .selectize-input")
        self._loc_selectize = self._loc_dropdown.locator(
            "> .selectize-dropdown-content"
        )
        self.loc = self.loc_container.locator(f"select#{id}")
        self.loc_choice_groups = self._loc_selectize.locator(
            "> .optgroup > .optgroup-header"
        )
        # Do not use `.option` class as we are not guaranteed to have it.
        # We are only guaranteed to have `data-value` attribute for each _option_
        self.loc_choices = self._loc_selectize.locator("[data-value]")
        self.loc_selected = self.loc_container.locator(f"select#{id} > option")

    def set(
        self,
        selected: str | list[str],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Sets the selected option(s) of the input selectize.

        Parameters
        ----------
        selected
            The value(s) of the selected option(s).
        timeout
            The maximum time to wait for the selection to be set. Defaults to `None`.
        """
        if isinstance(selected, str):
            selected = [selected]
        self._loc_events.click()
        for value in selected:
            self._loc_selectize.locator(f"[data-value='{value}']").click(
                timeout=timeout
            )
        self._loc_events.press("Escape")

    def expect_choices(
        self,
        # TODO-future; support patterns?
        choices: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the available options of the input selectize to be an exact match.

        Parameters
        ----------
        choices
            The expected choices of the input select.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        self._populate_dom()
        # Playwright doesn't like lists of size 0. Instead, check for empty locator
        if len(choices) == 0:
            playwright_expect(self.loc_choices).to_have_count(0, timeout=timeout)
            return

        expect_locator_values_in_list(
            page=self.page,
            loc_container=self._loc_selectize,
            el_type=self.page.locator("[data-value]"),
            arr_name="choices",
            arr=choices,
            key="data-value",
            timeout=timeout,
        )

    def expect_selected(
        self,
        value: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the selected option(s) of the input select to be an exact match.

        Parameters
        ----------
        value
            The expected value(s) of the selected option(s).
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        # Playwright doesn't like lists of size 0
        if isinstance(value, list) and len(value) == 0:
            playwright_expect(self.loc_selected).to_have_count(0, timeout=timeout)
            return

        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc,
            el_type=self.page.locator("> option"),
            arr_name="value",
            arr=value,
            key="value",
        )

    def _populate_dom(self, timeout: Timeout = None) -> None:
        """
        The click and Escape keypress is used to load the DOM elements
        """
        self._loc_events.click(timeout=timeout)
        _expect_style_to_have_value(
            self._loc_dropdown, "display", "block", timeout=timeout
        )
        self.page.locator("body").click(timeout=timeout)
        _expect_style_to_have_value(
            self._loc_dropdown, "display", "none", timeout=timeout
        )

    def expect_choice_groups(
        self,
        choice_groups: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the choice groups of the input select to be an exact match.

        Parameters
        ----------
        choice_groups
            The expected choice groups of the input select.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        self._populate_dom()
        # Playwright doesn't like lists of size 0. Instead, use `None`
        if len(choice_groups) == 0:
            playwright_expect(self.loc_choice_groups).to_have_count(0, timeout=timeout)
            return

        playwright_expect(self.loc_choice_groups).to_have_text(
            choice_groups, timeout=timeout
        )

    def expect_choice_labels(
        self,
        value: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the choice labels of the input selectize to be an exact match.

        Parameters
        ----------
        value
            The expected choice labels of the input select.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        self._populate_dom()
        # Playwright doesn't like lists of size 0. Instead, use `None`
        if len(value) == 0:
            playwright_expect(self.loc_choices).to_have_count(0, timeout=timeout)
            return
        playwright_expect(self.loc_choices).to_have_text(value, timeout=timeout)

    # multiple: bool = False,
    def expect_multiple(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expect the input selectize to allow multiple selections.

        Parameters
        ----------
        value
            Whether the input select allows multiple selections.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        if value:
            _expect_attribute_to_have_value(
                self.loc, "multiple", "multiple", timeout=timeout
            )
        else:
            _expect_attribute_to_have_value(self.loc, "multiple", None, timeout=timeout)


class InputSlider(_InputSliderBase):
    """Controller for :func:`shiny.ui.input_slider`."""

    loc_irs_label: Locator
    """
    Playwright `Locator` of the input slider label.
    """

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initialize the InputSlider object.

        Parameters
        ----------
        page
            The Playwright Page object.
        id
            The id of the input element.
        """
        super().__init__(page, id=id)
        self.loc_irs_label = self.loc_irs.locator("> .irs > .irs-single")

    def expect_value(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Asserts that the input element has the expected value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        playwright_expect(self.loc_irs_label).to_have_text(value, timeout=timeout)

    def set(
        self,
        value: str,
        *,
        max_err_values: int = 15,
        timeout: Timeout = None,
    ) -> None:
        """
        Set the value of the slider.

        Parameters
        ----------
        value
            The value to set the slider to.
        max_err_values
            The maximum number of error values to display if the value is not found. Defaults to 15.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        self._wait_for_container(timeout=timeout)

        handle = self.loc_irs.locator("> .irs-handle")
        handle_center = self._handle_center(handle, name="handle", timeout=timeout)
        grid_bb = self._grid_bb(timeout=timeout)

        self._set_helper(
            value=value,
            irs_label=self.loc_irs_label,
            handle_center=handle_center,
            grid_bb=grid_bb,
            start_x=grid_bb["x"],
            direction="right",
            max_err_values=max_err_values,
        )


class InputSliderRange(_InputSliderBase):
    """Controller for :func:`shiny.ui.input_slider` with a slider range."""

    loc_irs_label_from: Locator
    """
    Playwright `Locator` of the input slider label for the `from` handle.
    """
    loc_irs_label_to: Locator
    """
    Playwright `Locator` of the input slider label for the `to` handle.
    """

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initialize the InputSliderRange object.

        Parameters
        ----------
        page
            The Playwright Page object.
        id
            The id of the input element.
        """
        super().__init__(page, id=id)
        self.loc_irs_label_from = self.loc_irs.locator("> .irs > .irs-from")
        self.loc_irs_label_to = self.loc_irs.locator("> .irs > .irs-to")

    def expect_value(
        self,
        value: (
            typing.Tuple[PatternOrStr, PatternOrStr]
            | typing.Tuple[PatternOrStr, MISSING_TYPE]
            | typing.Tuple[MISSING_TYPE, PatternOrStr]
        ),
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        if all_missing(*value):
            raise ValueError("Both `value` tuple entries cannot be `MISSING_TYPE`")
        from_val = value[0]
        to_val = value[1]

        # TODO-future; Composable expectations
        if not_is_missing(from_val):
            playwright_expect(self.loc_irs_label_from).to_have_text(
                from_val, timeout=timeout
            )
        if not_is_missing(to_val):
            playwright_expect(self.loc_irs_label_to).to_have_text(
                to_val, timeout=timeout
            )

    def _set_fraction(
        self,
        handle: Locator,
        fraction: float,
        *,
        name: str,
        timeout: Timeout = None,
    ) -> None:
        if fraction > 1 or fraction < 0:
            raise ValueError("`fraction` must be between 0 and 1")

        handle_center = self._handle_center(handle, name=name, timeout=timeout)
        grid_bb = self._grid_bb(timeout=timeout)
        mouse = self.loc_container.page.mouse
        mouse.move(x=handle_center.get("x"), y=handle_center.get("y"))
        mouse.down()
        mouse.move(
            grid_bb.get("x") + (fraction * grid_bb.get("width")),
            handle_center.get("y"),
        )
        mouse.up()

    def set(
        self,
        value: (
            typing.Tuple[str, str]
            | typing.Tuple[str, MISSING_TYPE]
            | typing.Tuple[MISSING_TYPE, str]
        ),
        *,
        max_err_values: int = 15,
        timeout: Timeout = None,
    ) -> None:
        """
        Set the value of the slider.

        Parameters
        ----------
        value
            The value to set the slider to.
        max_err_values
            The maximum number of error values to display if the value is not found. Defaults to 15.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        if all_missing(*value):
            raise ValueError("Both `value` tuple entries cannot be `MISSING_TYPE`")

        value_from = value[0]
        value_to = value[1]
        self.loc_container.wait_for(state="visible", timeout=timeout)
        self.loc_container.scroll_into_view_if_needed(timeout=timeout)
        handle_from = self.loc_irs.locator("> .irs-handle.from")
        handle_to = self.loc_irs.locator("> .irs-handle.to")
        if not_is_missing(value_from):
            # Move `from` handle to the far left
            self._set_fraction(handle_from, 0, name="`from` handle", timeout=timeout)
        if not_is_missing(value_to):
            # Move `to` handle to the far right
            self._set_fraction(handle_to, 1, name="`to` handle", timeout=timeout)

        handle_center_from = self._handle_center(
            handle_from, name="`from` handle", timeout=timeout
        )
        grid_bb = self._grid_bb(timeout=timeout)

        # Handles are [possibly] now at their respective extreme value
        # Now let's move them towards the other end until we find the corresponding str

        if not_is_missing(value_from):
            self._set_helper(
                value=value_from,
                irs_label=self.loc_irs_label_from,
                handle_center=handle_center_from,
                grid_bb=grid_bb,
                # Start at the far left
                start_x=grid_bb["x"],
                # And move to the right
                direction="right",
                max_err_values=max_err_values,
            )

        handle_center_to = self._handle_center(
            handle_to, name="`to` handle", timeout=timeout
        )
        if not_is_missing(value_to):
            self._set_helper(
                value=value_to,
                irs_label=self.loc_irs_label_to,
                handle_center=handle_center_to,
                grid_bb=grid_bb,
                # Start at the far right
                start_x=grid_bb["x"] + grid_bb["width"],
                # And move to the left
                direction="left",
                max_err_values=max_err_values,
            )



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.playwright.expect import expect_to_have_class
from shiny.run import ShinyAppProc


def get_value_box_bg_color(value_box: controller.ValueBox) -> str:
    value_box_bg_color = value_box.loc_container.evaluate(
        "el => window.getComputedStyle(el).getPropertyValue('background-color');"
    )
    return value_box_bg_color


def get_value_box_fg_color(value_box: controller.ValueBox) -> str:
    value_box_fg_color = value_box.loc_container.evaluate(
        "el => window.getComputedStyle(el).getPropertyValue('color');"
    )
    return value_box_fg_color


def get_title_tag_name(value_box: controller.ValueBox) -> str:
    title_tag_name = (
        value_box.loc_title.locator("*")
        .nth(0)
        .evaluate("el => el.tagName.toLowerCase()")
    )
    return title_tag_name


def get_value_tag_name(value_box: controller.ValueBox) -> str:
    value_tag_name = (
        value_box.loc.locator("*").nth(0).evaluate("el => el.tagName.toLowerCase()")
    )
    return value_tag_name


"""
For each value box we want to test
Layout and Positioning(e.g., showcase-top-right, showcase-left-center, showcase-bottom).
Title and Value Elements: The tests assert the tag names used for the title and value elements within each ValueBox (e.g., <span>, <p>, <h1>, <h3>, <h5>).
Fullscreen Availability and State: The tests check whether the fullscreen feature is available for a particular ValueBox and verify its initial state (fullscreen or not). For ValueBoxes with fullscreen support, the tests open and close the fullscreen mode and assert the expected behavior.
Background and Foreground Colors: The tests assert the background and foreground (text) colors applied to the ValueBox components.
Height: In some cases, the tests expect a specific height value for the ValueBox.
Content: The tests verify the expected title and value text displayed within each ValueBox. Additionally, for ValueBoxes with fullscreen support, the tests check the content displayed in fullscreen mode.
"""


def test_valuebox(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    value_box1 = controller.ValueBox(page, "valuebox1")
    value_box1.expect_height(None)
    # verify showcase layout uses 'showcase_top_right()'
    expect_to_have_class(value_box1.loc_container, "showcase-top-right")
    assert get_title_tag_name(value_box1) == "span"
    value_box1.expect_title("Red Color theme w/ Fullscreen")
    value_box1.expect_value("Showcase top right")
    assert get_value_tag_name(value_box1) == "h1"
    assert get_value_box_bg_color(value_box1) == "rgb(193, 0, 0)"
    assert get_value_box_fg_color(value_box1) == "rgb(255, 255, 255)"
    value_box1.expect_full_screen_available(True)
    value_box1.expect_full_screen(False)
    value_box1.set_full_screen(True)
    value_box1.expect_full_screen(True)
    value_box1.expect_body(["Inside the fullscreen"])
    value_box1.set_full_screen(False)
    value_box1.expect_full_screen(False)

    value_box2 = controller.ValueBox(page, "valuebox2")
    value_box2.expect_height(None)
    # verify showcase layout uses 'showcase_left_center()'
    expect_to_have_class(value_box2.loc_container, "showcase-left-center")
    assert get_title_tag_name(value_box2) == "p"
    value_box2.expect_title("Primary theme w/o Fullscreen")
    value_box2.expect_value("Showcase left center")
    assert get_value_tag_name(value_box2) == "h5"
    assert get_value_box_bg_color(value_box2) == "rgb(0, 123, 194)"
    value_box2.expect_full_screen_available(False)

    value_box3 = controller.ValueBox(page, "valuebox3")
    value_box3.expect_height(None)
    # verify showcase layout uses 'showcase_bottom()'
    expect_to_have_class(value_box3.loc_container, "showcase-bottom")
    assert get_title_tag_name(value_box3) == "span"
    value_box3.expect_title("No theme w/ Fullscreen")
    value_box3.expect_value("Showcase bottom")
    assert get_value_tag_name(value_box3) == "h3"
    value_box3.expect_full_screen_available(True)
    assert get_value_box_bg_color(value_box3) == "rgb(255, 255, 255)"

    value_box4 = controller.ValueBox(page, "valuebox4")
    value_box4.expect_height(None)
    value_box4.expect_title("No showcase - w/o Fullscreen (default)")
    value_box4.expect_value("No theme - only defaults")
    assert get_title_tag_name(value_box4) == "p"
    assert get_title_tag_name(value_box4) == "p"
    value_box4.expect_full_screen_available(False)
    assert get_value_box_bg_color(value_box4) == "rgb(255, 255, 255)"

    value_box5 = controller.ValueBox(page, "valuebox5")
    value_box5.expect_height("500px")
    assert get_title_tag_name(value_box5) == "p"
    value_box5.expect_title("No showcase w/ showcase layout")
    value_box5.expect_value("Red text - fill is False")
    assert get_value_tag_name(value_box5) == "p"
    value_box5.expect_full_screen_available(False)
    assert get_value_box_bg_color(value_box5) == "rgb(255, 255, 255)"
    assert get_value_box_fg_color(value_box5) == "rgb(193, 0, 0)"



from model_plots import *  # model plots and cards

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.panel_title(ui.h2("Model Dashboard")),
    ui.markdown("Using `ui.layout_columns()` for the layout."),
    ui.layout_columns(
        card_loss,
        card_acc,
        card_feat,
        col_widths={"sm": (5, 7, 12)},
        # row_heights=(2, 3),
        # height="700px",
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.plot
    def loss_over_time():
        return plot_loss_over_time()

    @render.plot
    def accuracy_over_time():
        return plot_accuracy_over_time()

    @render.plot
    def feature_importance():
        return plot_feature_importance()


app = App(app_ui, server)



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_password")


def test_input_password_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    password = controller.InputPassword(page, "password")
    password.expect.to_have_value("")
    expect(password.loc).to_have_value("")

    expect(password.loc_label).to_have_text("Password:")

    password.expect_label("Password:")
    password.expect_value("")
    password.expect_width(None)
    password.expect_placeholder(None)

    password_value = "test password"
    password.set(password_value)

    controller.InputActionButton(page, "go").click()
    password.expect_value(password_value)
    controller.OutputTextVerbatim(page, "value").expect.to_have_text(password_value)
